# coding=utf-8
"""
remind.py - Sopel Reminder Plugin
Copyright 2011, Sean B. Palmer, inamidst.com
Copyright 2019, dgw, technobabbl.es
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
from datetime import datetime
import io  # don't use `codecs` for loading the DB; it will split lines on some IRC formatting
import logging
import os
import re
import time

import pytz

from sopel import plugin, tools
from sopel.tools.time import format_time, get_timezone, validate_timezone


LOGGER = logging.getLogger(__name__)


def get_filename(bot):
    """Get the remind database's filename

    :param bot: instance of Sopel
    :type bot: :class:`sopel.bot.Sopel`
    :return: the remind database's filename
    :rtype: str

    The remind database filename is based on the bot's nick and its
    configured ``core.host``, and it is located in the ``bot``'s ``homedir``.
    """
    name = bot.config.basename + '.reminders.db'
    return os.path.join(bot.config.core.homedir, name)


def load_database(filename):
    """Load the remind database from a file

    :param str filename: absolute path to the remind database file
    :return: a :class:`dict` of reminders stored by timestamp
    :rtype: dict

    The remind database is a plain text file (utf-8 encoded) with tab-separated
    columns of data: time, channel, nick, and message. This function reads this
    file and outputs a :class:`dict` where keys are the timestamps of the
    reminders, and values are list of 3-value tuple of reminder data:
    ``(channel, nick, message)``.

    .. note::

        This function ignores microseconds from the timestamp, if any, meaning
        that ``523549800.245`` will be read as ``523549800``.

    .. note::

        If ``filename`` is not an existing file, this function returns an
        empty :class:`dict`.

    """
    if not os.path.isfile(filename):
        # no file to read
        return {}

    data = {}
    with io.open(filename, 'r', encoding='utf-8') as database:
        for line in database:
            unixtime, channel, nick, message = line.split('\t', 3)
            message = message.rstrip('\n')
            timestamp = int(float(unixtime))  # ignore microseconds
            reminder = (channel, nick, message)
            try:
                data[timestamp].append(reminder)
            except KeyError:
                data[timestamp] = [reminder]
    return data


def dump_database(filename, data):
    """Dump the remind database into a file

    :param str filename: absolute path to the remind database file
    :param dict data: remind database to dump

    The remind database is dumped into a plain text file (utf-8 encoded) as
    tab-separated columns of data: unixtime, channel, nick, and message.

    If the file does not exist, it is created.
    """
    with io.open(filename, 'w', encoding='utf-8') as database:
        for unixtime, reminders in tools.iteritems(data):
            for channel, nick, message in reminders:
                line = '%s\t%s\t%s\t%s\n' % (unixtime, channel, nick, message)
                database.write(line)


def create_reminder(bot, trigger, duration, message):
    """Create a reminder into the ``bot``'s database and reply to the sender

    :param bot: the bot's instance
    :type bot: :class:`~sopel.bot.SopelWrapper`
    :param trigger: the object that triggered the call
    :type trigger: :class:`~sopel.trigger.Trigger`
    :param int duration: duration from now, in seconds, until ``message``
                         must be reminded
    :param str message: message to be reminded
    :return: the reminder's timestamp
    :rtype: :class:`int`
    """
    timestamp = int(time.time()) + duration
    reminder = (trigger.sender, trigger.nick, message)
    try:
        bot.rdb[timestamp].append(reminder)
    except KeyError:
        bot.rdb[timestamp] = [reminder]

    dump_database(bot.rfn, bot.rdb)
    return timestamp


def setup(bot):
    """Load the remind database"""
    bot.rfn = get_filename(bot)

    # Pre-7.0 migration logic. Remove in 8.0 or 9.0.
    old = bot.nick + '-' + bot.config.core.host + '.reminders.db'
    old = os.path.join(bot.config.core.homedir, old)
    if os.path.isfile(old):
        LOGGER.info("Attempting to migrate old 'remind' database {}..."
                    .format(old))
        try:
            os.rename(old, bot.rfn)
        except OSError:
            LOGGER.error("Migration failed!")
            LOGGER.error("Old filename: {}".format(old))
            LOGGER.error("New filename: {}".format(bot.rfn))
            LOGGER.error(
                "See https://sopel.chat/usage/installing/upgrading-to-sopel-7/#reminder-db-migration")
        else:
            LOGGER.info("Migration finished!")
    # End migration logic

    bot.rdb = load_database(bot.rfn)


def shutdown(bot):
    """Dump the remind database before shutdown"""
    dump_database(bot.rfn, bot.rdb)
    bot.rdb = {}
    del bot.rfn
    del bot.rdb


@plugin.interval(2.5)
def remind_monitoring(bot):
    """Check for reminder"""
    now = int(time.time())
    unixtimes = [int(key) for key in bot.rdb]
    oldtimes = [t for t in unixtimes if t <= now]
    if oldtimes:
        for oldtime in oldtimes:
            for (channel, nick, message) in bot.rdb[oldtime]:
                if message:
                    bot.say(nick + ': ' + message, channel)
                else:
                    bot.say(nick + '!', channel)
            del bot.rdb[oldtime]
        dump_database(bot.rfn, bot.rdb)


SCALING = collections.OrderedDict([
    ('years', 365.25 * 24 * 3600),
    ('year', 365.25 * 24 * 3600),
    ('yrs', 365.25 * 24 * 3600),
    ('y', 365.25 * 24 * 3600),

    ('months', 29.53059 * 24 * 3600),
    ('month', 29.53059 * 24 * 3600),
    ('mo', 29.53059 * 24 * 3600),

    ('weeks', 7 * 24 * 3600),
    ('week', 7 * 24 * 3600),
    ('wks', 7 * 24 * 3600),
    ('wk', 7 * 24 * 3600),
    ('w', 7 * 24 * 3600),

    ('days', 24 * 3600),
    ('day', 24 * 3600),
    ('d', 24 * 3600),

    ('hours', 3600),
    ('hour', 3600),
    ('hrs', 3600),
    ('hr', 3600),
    ('h', 3600),

    ('minutes', 60),
    ('minute', 60),
    ('mins', 60),
    ('min', 60),
    ('m', 60),

    ('seconds', 1),
    ('second', 1),
    ('secs', 1),
    ('sec', 1),
    ('s', 1),
])

PERIODS = '|'.join(SCALING.keys())


@plugin.command('in')
@plugin.example('.in 3h45m Go to class')
def remind_in(bot, trigger):
    """Gives you a reminder in the given amount of time."""
    if not trigger.group(2):
        bot.reply("Missing arguments for reminder command.")
        return plugin.NOLIMIT
    if trigger.group(3) and not trigger.group(4):
        bot.reply("No message given for reminder.")
        return plugin.NOLIMIT
    duration = 0
    message = filter(None, re.split(r'(\d+(?:\.\d+)? ?(?:(?i)' + PERIODS + ')) ?',
                                    trigger.group(2))[1:])
    reminder = ''
    stop = False
    for piece in message:
        grp = re.match(r'(\d+(?:\.\d+)?) ?(.*) ?', piece)
        if grp and not stop:
            length = float(grp.group(1))
            factor = SCALING.get(grp.group(2).lower(), 60)
            duration += length * factor
        else:
            reminder = reminder + piece
            stop = True
    if duration == 0:
        bot.reply("Sorry, didn't understand the input.")
        return plugin.NOLIMIT

    if duration % 1:
        duration = int(duration) + 1
    else:
        duration = int(duration)
    timezone = get_timezone(
        bot.db, bot.config, None, trigger.nick, trigger.sender)
    timestamp = create_reminder(bot, trigger, duration, reminder)

    if duration >= 60:
        human_time = format_time(
            bot.db,
            bot.config,
            timezone,
            trigger.nick,
            trigger.sender,
            datetime.utcfromtimestamp(timestamp))
        bot.reply('Okay, will remind at %s' % human_time)
    else:
        bot.reply('Okay, will remind in %s secs' % duration)


REGEX_AT = re.compile(
    # hours:minutes
    r'(?P<hours>\d+):(?P<minutes>\d+)'
    # optional seconds
    r'(?::(?P<seconds>\d+))?'
    # optional timezone
    r'(?P<tz>[^\s\d]+)?'
    # optional date (start)
    r'(?:\s+'
    # - date 1 (at least one digit)
    r'(?P<date1>\d{1,4})'
    # - separator (one character)
    r'(?P<sep>[./-])'
    # - date 2 (at least one digit)
    r'(?P<date2>\d{1,4})'
    # - optional sep + date 3 (at least one digit)
    r'(?:(?P=sep)(?P<date3>\d{1,4}))?'
    r')?'  # (end)
    # at least one space + message
    r'\s+(?P<message>.*)'
)


class TimeReminder(object):
    """Time reminder for the ``at`` command"""
    def __init__(self,
                 hour,
                 minute,
                 second,
                 timezone,
                 date1,
                 date2,
                 date3,
                 message):
        self.hour = hour
        self.minute = minute
        self.second = second
        self.timezone = pytz.timezone(timezone)
        self.message = message

        year = None
        month = None
        day = None

        if date1 and date2 and date3:
            if len(date1) == 4:
                # YYYY-mm-dd
                year = int(date1)
                month = int(date2)
                day = int(date3)
            else:
                # dd-mm-YYYY or dd/mm/YY
                year = int(date3)
                month = int(date2)
                day = int(date1)
        elif date1 and date2:
            if len(date1) == 4:
                # YYYY-mm
                year = int(date1)
                month = int(date2)
            elif len(date2) == 4:
                # mm-YYYY
                year = int(date2)
                month = int(date1)
            else:
                # dd/mm
                month = int(date2)
                day = int(date1)

        self.year = year
        self.month = month
        self.day = day

    def __eq__(self, other):
        return all(
            getattr(self, attr) == getattr(other, attr, None)
            for attr in [
                'hour',
                'minute',
                'second',
                'timezone',
                'year',
                'month',
                'day',
                'message',
            ]
        )

    def __ne__(self, other):
        return any(
            getattr(self, attr) != getattr(other, attr, None)
            for attr in [
                'hour',
                'minute',
                'second',
                'timezone',
                'year',
                'month',
                'day',
                'message',
            ]
        )

    # Mutable objects probably shouldn't be made hashable
    # https://docs.python.org/3/reference/datamodel.html#object.__hash__
    __hash__ = None

    def get_duration(self, today=None):
        """Get the duration between the reminder and ``today``

        :param today: aware datetime to compare to; defaults to current time
        :type today: aware :class:`datetime.datetime`
        :return: The duration, in second, between ``today`` and the reminder.
        :rtype: :class:`int`

        This method returns the number of seconds given by the
        :class:`datetime.timedelta` obtained by the difference between the
        reminder and ``today``.

        If the delta between the reminder and ``today`` is negative, Python
        will represent it as a negative number of days, and a positive number
        of seconds: since it returns the number of seconds, any past reminder
        will be transformed into a future reminder the next day.

        .. seealso::

            The :mod:`datetime` built-in module's documentation explains what
            is an "aware" datetime.

        """
        if not today:
            today = datetime.now(self.timezone)
        else:
            today = today.astimezone(self.timezone)

        year = self.year if self.year is not None else today.year
        month = self.month if self.month is not None else today.month
        day = self.day if self.day is not None else today.day

        at_time = datetime(
            year, month, day,
            self.hour, self.minute, self.second,
            tzinfo=today.tzinfo)

        timediff = at_time - today
        duration = timediff.seconds

        if timediff.days > 0:
            duration = duration + (86400 * timediff.days)

        return duration


def parse_regex_match(match, default_timezone=None):
    """Parse a time reminder from ``match``

    :param match: :obj:`~.REGEX_AT`'s matching result
    :param default_timezone: timezone used when ``match`` doesn't have one;
                             defaults to ``UTC``
    :rtype: :class:`TimeReminder`
    """
    try:
        # Removing the `or` clause will BREAK the fallback to default_timezone!
        # We need some invalid value other than None to trigger the ValueError.
        # validate_timezone(None) excepting would be easier, but it doesn't.
        timezone = validate_timezone(match.group('tz') or '')
    except ValueError:
        timezone = default_timezone or 'UTC'

    return TimeReminder(
        int(match.group('hours')),
        int(match.group('minutes')),
        int(match.group('seconds') or '0'),
        timezone,
        match.group('date1'),
        match.group('date2'),
        match.group('date3'),
        match.group('message')
    )


@plugin.command('at')
@plugin.example('.at 13:47 Do your homework!', user_help=True)
@plugin.example('.at 03:14:07 2038-01-19 End of signed 32-bit int timestamp',
                user_help=True)
@plugin.example('.at 00:01 25/12 Open your gift!', user_help=True)
def remind_at(bot, trigger):
    """Gives you a reminder at the given time.

    Takes ``hh:mm:ssTimezone Date message`` where seconds, Timezone, and Date
    are optional.

    Timezone is any timezone Sopel takes elsewhere; the best choices are those
    from the tzdb; a list of valid options is available at
    <https://sopel.chat/tz>.

    The Date can be expressed in one of these formats: YYYY-mm-dd, dd-mm-YYYY,
    dd-mm-YY, or dd-mm. The separator can be ``.``, ``-``, or ``/``.
    """
    if not trigger.group(2):
        bot.reply("No arguments given for reminder command.")
        return plugin.NOLIMIT
    if trigger.group(3) and not trigger.group(4):
        bot.reply("No message given for reminder.")
        return plugin.NOLIMIT

    match = REGEX_AT.match(trigger.group(2))
    if not match:
        bot.reply("Sorry, but I didn't understand your input.")
        return plugin.NOLIMIT

    default_timezone = get_timezone(bot.db, bot.config, None,
                                    trigger.nick, trigger.sender)

    reminder = parse_regex_match(match, default_timezone)

    try:
        duration = reminder.get_duration()
    except ValueError as error:
        bot.reply(
            "Sorry, but I didn't understand your input: %s" % str(error))
        return plugin.NOLIMIT

    # save reminder
    timestamp = create_reminder(bot, trigger, duration, reminder.message)

    if duration >= 60:
        human_time = format_time(
            bot.db,
            bot.config,
            reminder.timezone.zone,
            trigger.nick,
            trigger.sender,
            datetime.utcfromtimestamp(timestamp))
        bot.reply('Okay, will remind at %s' % human_time)
    else:
        bot.reply('Okay, will remind in %s secs' % duration)


@plugin.command('reminders')
@plugin.example('.reminders forget *', user_help=True)
@plugin.example('.reminders count #channel', user_help=True)
@plugin.example('.reminders count', user_help=True)
def manage_reminders(bot, trigger):
    """Count or forget your reminders in the current channel.

    Use a subcommand "count" (default) or "forget". The second argument is
    optional and can be either a channel name, your nick, or * (for all).
    """
    owner = trigger.nick
    action = trigger.group(3) or trigger.sender
    target = trigger.group(4)

    if action not in ['count', 'forget'] and not target:
        # assume `.reminders` or `.reminders #channel`
        # in that case, invalid action will just count 0 reminder
        action, target = 'count', action

    if action == 'count':
        tpl = 'You have {count} reminders for all channels.'
        nick_reminders = (
            (timestamp, channel, nick, message)
            for timestamp, reminders in bot.rdb.items()
            for channel, nick, message in reminders
            if nick == owner
        )
        if target and target != '*':
            tpl = 'You have {count} reminders in {target}.'
            nick_reminders = (
                (timestamp, channel, nick, message)
                for timestamp, channel, nick, message in nick_reminders
                if channel == target
            )

        count = sum(1 for __ in nick_reminders)

        if target == owner:
            target = 'private'

        bot.reply(tpl.format(count=count, target=target))

    elif action == 'forget':
        bot.rdb = {
            timestamp: [
                (channel, nick, message)
                for channel, nick, message in reminders
                if not (
                    nick == owner
                    and (target == '*' or target == channel)
                )
            ]
            for timestamp, reminders in bot.rdb.items()
        }
        dump_database(bot.rfn, bot.rdb)

        if not target or target == '*':
            bot.reply('I forgot all your reminders.')
        elif target == owner:
            bot.reply('I forgot your private reminders.')
        else:
            bot.reply('I forgot your reminders in %s' % target)

    else:
        bot.reply(
            'Unrecognized action. '
            'Usage: {}reminders [count|forget [nickname|channel|*]]'
            .format(bot.config.core.help_prefix))



# coding=utf-8
"""
url.py - Sopel URL Title Plugin
Copyright 2010-2011, Michael Yanovich (yanovich.net) & Kenneth Sham
Copyright 2012-2013, Elsie Powell
Copyright 2013, Lior Ramati <firerogue517@gmail.com>
Copyright 2014, Elad Alfassa <elad@fedoraproject.org>
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import ipaddress
import logging
import re

import dns.resolver
import requests
from urllib3.exceptions import LocationValueError

from sopel import plugin, tools
from sopel.config import types
from sopel.tools import web

# Python3 vs Python2
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

LOGGER = logging.getLogger(__name__)
USER_AGENT = (
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
    'AppleWebKit/537.36 (KHTML, like Gecko) '
    'Chrome/78.0.3904.108 Safari/537.36'
)
DEFAULT_HEADERS = {
    'User-Agent': USER_AGENT,
    'Accept': 'text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8',
    'Accept-Language': 'en,en-US;q=0,5',
}
# These are used to clean up the title tag before actually parsing it. Not the
# world's best way to do this, but it'll do for now.
TITLE_TAG_DATA = re.compile('<(/?)title( [^>]+)?>', re.IGNORECASE)
QUOTED_TITLE = re.compile('[\'"]<title>[\'"]', re.IGNORECASE)
# This is another regex that presumably does something important.
RE_DCC = re.compile(r'(?i)dcc\ssend')
# This sets the maximum number of bytes that should be read in order to find
# the title. We don't want it too high, or a link to a big file/stream will
# just keep downloading until there's no more memory. 640k ought to be enough
# for anybody, but the modern web begs to differ.
MAX_BYTES = 655360 * 2


class UrlSection(types.StaticSection):
    enable_auto_title = types.BooleanAttribute(
        'enable_auto_title', default=True)
    """Enable auto-title (enabled by default)"""
    # TODO some validation rules maybe?
    exclude = types.ListAttribute('exclude')
    """A list of regular expressions to match URLs for which the title should not be shown."""
    exclusion_char = types.ValidatedAttribute('exclusion_char', default='!')
    """A character (or string) which, when immediately preceding a URL, will stop that URL's title from being shown."""
    shorten_url_length = types.ValidatedAttribute(
        'shorten_url_length', int, default=0)
    """If greater than 0, the title fetcher will include a TinyURL version of links longer than this many characters."""
    enable_private_resolution = types.BooleanAttribute(
        'enable_private_resolution', default=False)
    """Enable URL lookups for RFC1918 addresses"""
    enable_dns_resolution = types.BooleanAttribute(
        'enable_dns_resolution', default=False)
    """Enable DNS resolution for all domains to validate if there are RFC1918 resolutions"""


def configure(config):
    """
    | name | example | purpose |
    | ---- | ------- | ------- |
    | enable_auto_title | yes | Enable auto-title. |
    | exclude | https?://git\\\\.io/.* | A list of regular expressions for URLs for which the title should not be shown. |
    | exclusion\\_char | ! | A character (or string) which, when immediately preceding a URL, will stop the URL's title from being shown. |
    | shorten\\_url\\_length | 72 | If greater than 0, the title fetcher will include a TinyURL version of links longer than this many characters. |
    | enable\\_private\\_resolution | False | Enable URL lookups for RFC1918 addresses. |
    | enable\\_dns\\_resolution | False | Enable DNS resolution for all domains to validate if there are RFC1918 resolutions. |
    """
    config.define_section('url', UrlSection)
    config.url.configure_setting(
        'enable_auto_title',
        'Enable auto-title?'
    )
    config.url.configure_setting(
        'exclude',
        'Enter regular expressions for each URL you would like to exclude.'
    )
    config.url.configure_setting(
        'exclusion_char',
        'Enter a character which can be prefixed to suppress URL titling'
    )
    config.url.configure_setting(
        'shorten_url_length',
        'Enter how many characters a URL should be before the bot puts a'
        ' shorter version of the URL in the title as a TinyURL link'
        ' (0 to disable)'
    )
    config.url.configure_setting(
        'enable_private_resolution',
        'Enable URL lookups for RFC1918 addresses?'
    )
    config.url.configure_setting(
        'enable_dns_resolution',
        'Enable DNS resolution for all domains to validate if there are RFC1918 resolutions?'
    )


def setup(bot):
    bot.config.define_section('url', UrlSection)

    if bot.config.url.exclude:
        regexes = [re.compile(s) for s in bot.config.url.exclude]
    else:
        regexes = []

    # We're keeping these in their own list, rather than putting then in the
    # callbacks list because 1, it's easier to deal with plugins that are still
    # using this list, and not the newer callbacks list and 2, having a lambda
    # just to pass is kinda ugly.
    if 'url_exclude' not in bot.memory:
        bot.memory['url_exclude'] = regexes
    else:
        exclude = bot.memory['url_exclude']
        if regexes:
            exclude.extend(regexes)
        bot.memory['url_exclude'] = exclude

    # Ensure last_seen_url is in memory
    if 'last_seen_url' not in bot.memory:
        bot.memory['last_seen_url'] = tools.SopelIdentifierMemory()

    # Initialize shortened_urls as a dict if it doesn't exist.
    if 'shortened_urls' not in bot.memory:
        bot.memory['shortened_urls'] = tools.SopelMemory()


def shutdown(bot):
    # Unset `url_exclude` and `last_seen_url`, but not `shortened_urls`;
    # clearing `shortened_urls` will increase API calls. Leaving it in memory
    # should not lead to unexpected behavior.
    for key in ['url_exclude', 'last_seen_url']:
        try:
            del bot.memory[key]
        except KeyError:
            pass


@plugin.command('urlexclude', 'urlpexclude', 'urlban', 'urlpban')
@plugin.example('.urlpexclude example\\.com/\\w+', user_help=True)
@plugin.example('.urlexclude example.com/path', user_help=True)
@plugin.output_prefix('[url] ')
def url_ban(bot, trigger):
    """Exclude a URL from auto title.

    Use ``urlpexclude`` to exclude a pattern instead of a URL.
    """
    url = trigger.group(2)

    if not url:
        bot.reply('This command requires a URL to exclude.')
        return

    if trigger.group(1) in ['urlpexclude', 'urlpban']:
        # validate regex pattern
        try:
            re.compile(url)
        except re.error as err:
            bot.reply('Invalid regex pattern: %s' % err)
            return
    else:
        # escape the URL to ensure a valid pattern
        url = re.escape(url)

    patterns = bot.settings.url.exclude

    if url in patterns:
        bot.reply('This URL is already excluded from auto title.')
        return

    # update settings
    patterns.append(url)
    bot.settings.url.exclude = patterns  # set the config option
    bot.settings.save()
    LOGGER.info('%s excluded the URL pattern "%s"', trigger.nick, url)

    # re-compile
    bot.memory['url_exclude'] = [re.compile(s) for s in patterns]

    # tell the user
    bot.reply('This URL is now excluded from auto title.')


@plugin.command('urlallow', 'urlpallow', 'urlunban', 'urlpunban')
@plugin.example('.urlpallow example\\.com/\\w+', user_help=True)
@plugin.example('.urlallow example.com/path', user_help=True)
@plugin.output_prefix('[url] ')
def url_unban(bot, trigger):
    """Allow a URL for auto title.

    Use ``urlpallow`` to allow a pattern instead of a URL.
    """
    url = trigger.group(2)

    if not url:
        bot.reply('This command requires a URL to allow.')
        return

    if trigger.group(1) in ['urlpallow', 'urlpunban']:
        # validate regex pattern
        try:
            re.compile(url)
        except re.error as err:
            bot.reply('Invalid regex pattern: %s' % err)
            return
    else:
        # escape the URL to ensure a valid pattern
        url = re.escape(url)

    patterns = bot.settings.url.exclude

    if url not in patterns:
        bot.reply('This URL was not excluded from auto title.')
        return

    # update settings
    patterns.remove(url)
    bot.settings.url.exclude = patterns  # set the config option
    bot.settings.save()
    LOGGER.info('%s allowed the URL pattern "%s"', trigger.nick, url)

    # re-compile
    bot.memory['url_exclude'] = [re.compile(s) for s in patterns]

    # tell the user
    bot.reply('This URL is not excluded from auto title anymore.')


@plugin.command('title')
@plugin.example(
    '.title https://www.google.com',
    'Google | www.google.com',
    online=True, vcr=True)
@plugin.output_prefix('[url] ')
def title_command(bot, trigger):
    """
    Show the title or URL information for the given URL, or the last URL seen
    in this channel.
    """
    if not trigger.group(2):
        if trigger.sender not in bot.memory['last_seen_url']:
            return
        matched = check_callbacks(
            bot, bot.memory['last_seen_url'][trigger.sender])
        if matched:
            return
        else:
            urls = [bot.memory['last_seen_url'][trigger.sender]]
    else:
        urls = list(  # needs to be a list so len() can be checked later
            web.search_urls(
                trigger,
                exclusion_char=bot.config.url.exclusion_char
            )
        )

    result_count = 0
    for url, title, domain, tinyurl in process_urls(bot, trigger, urls):
        message = '%s | %s' % (title, domain)
        if tinyurl:
            message += ' ( %s )' % tinyurl
        bot.reply(message)
        bot.memory['last_seen_url'][trigger.sender] = url
        result_count += 1

    expected_count = len(urls)
    if result_count < expected_count:
        if expected_count == 1:
            bot.reply("Sorry, fetching that title failed. Make sure the site is working.")
        elif result_count == 0:
            bot.reply("Sorry, I couldn't fetch titles for any of those.")
        else:
            bot.reply("I couldn't get all of the titles, but I fetched what I could!")


@plugin.rule(r'(?u).*(https?://\S+).*')
@plugin.output_prefix('[url] ')
def title_auto(bot, trigger):
    """
    Automatically show titles for URLs. For shortened URLs/redirects, find
    where the URL redirects to and show the title for that (or call a function
    from another plugin to give more information).
    """
    # Enabled or disabled by feature flag
    if not bot.settings.url.enable_auto_title:
        return

    # Avoid fetching links from another command
    if re.match(bot.config.core.prefix + r'\S+', trigger):
        return

    # Avoid fetching known malicious links
    if 'safety_cache' in bot.memory and trigger in bot.memory['safety_cache']:
        if bot.memory['safety_cache'][trigger]['positives'] > 1:
            return

    urls = web.search_urls(
        trigger, exclusion_char=bot.config.url.exclusion_char, clean=True)

    for url, title, domain, tinyurl in process_urls(bot, trigger, urls):
        message = '%s | %s' % (title, domain)
        if tinyurl:
            message += ' ( %s )' % tinyurl
        # Guard against responding to other instances of this bot.
        if message != trigger:
            bot.say(message)
            bot.memory['last_seen_url'][trigger.sender] = url


def process_urls(bot, trigger, urls):
    """
    For each URL in the list, ensure that it isn't handled by another plugin.
    If not, find where it redirects to, if anywhere. If that redirected URL
    should be handled by another plugin, dispatch the callback for it.
    Return a list of (title, hostname) tuples for each URL which is not handled
    by another plugin.
    """
    shorten_url_length = bot.config.url.shorten_url_length
    for url in urls:
        # Exclude URLs that start with the exclusion char
        if url.startswith(bot.config.url.exclusion_char):
            continue

        # Check the URL does not match an existing URL callback
        if check_callbacks(bot, url):
            continue

        # Prevent private addresses from being queried if enable_private_resolution is False
        if not bot.config.url.enable_private_resolution:
            parsed = urlparse(url)
            # Check if it's an address like http://192.168.1.1
            try:
                if ipaddress.ip_address(parsed.hostname).is_private or ipaddress.ip_address(parsed.hostname).is_loopback:
                    LOGGER.debug('Ignoring private URL: %s', url)
                    continue
            except ValueError:
                pass

            # Check if domains are RFC1918 addresses if enable_dns_resolutions is set
            if bot.config.url.enable_dns_resolution:
                private = False
                for result in dns.resolver.query(parsed.hostname):
                    if ipaddress.ip_address(result).is_private or ipaddress.ip_address(parsed.hostname).is_loopback:
                        private = True
                        break
                if private:
                    LOGGER.debug('Ignoring private URL: %s', url)
                    continue

        # Call the URL to get a title, if possible
        title = find_title(url)
        if not title:
            # No title found: don't handle this URL
            LOGGER.warning('No title found; ignoring URL: %s', url)
            continue

        # If the URL is over bot.config.url.shorten_url_length, shorten the URL
        tinyurl = None
        if (shorten_url_length > 0) and (len(url) > shorten_url_length):
            tinyurl = get_or_create_shorturl(bot, url)

        yield (url, title, get_hostname(url), tinyurl)


def check_callbacks(bot, url):
    """Check if ``url`` is excluded or matches any URL callback patterns.

    :param bot: Sopel instance
    :param str url: URL to check
    :return: True if ``url`` is excluded or matches any URL callback pattern

    This function looks at the ``bot.memory`` for ``url_exclude`` patterns and
    it returns ``True`` if any matches the given ``url``. Otherwise, it looks
    at the ``bot``'s URL callback patterns, and it returns ``True`` if any
    matches, ``False`` otherwise.

    .. seealso::

        The :func:`~sopel.modules.url.setup` function that defines the
        ``url_exclude`` in ``bot.memory``.

    .. versionchanged:: 7.0

        This function **does not** trigger URL callbacks anymore when ``url``
        matches a pattern.

    """
    # Check if it matches the exclusion list first
    matched = any(regex.search(url) for regex in bot.memory['url_exclude'])
    return (
        matched or
        any(bot.search_url_callbacks(url)) or
        bot.rules.check_url_callback(bot, url)
    )


def find_title(url, verify=True):
    """Return the title for the given URL."""
    try:
        response = requests.get(url, stream=True, verify=verify,
                                headers=DEFAULT_HEADERS)
        content = b''
        for byte in response.iter_content(chunk_size=512):
            content += byte
            if b'</title>' in content or len(content) > MAX_BYTES:
                break
        content = content.decode('utf-8', errors='ignore')
        # Need to close the connection because we have not read all
        # the data
        response.close()
    except requests.exceptions.ConnectionError:
        LOGGER.exception('Unable to reach URL: %s', url)
        return None
    except (
        requests.exceptions.InvalidURL,  # e.g. http:///
        UnicodeError,  # e.g. http://.example.com (urllib3<1.26)
        LocationValueError,  # e.g. http://.example.com (urllib3>=1.26)
    ):
        LOGGER.debug('Invalid URL: %s', url)
        return None

    # Some cleanup that I don't really grok, but was in the original, so
    # we'll keep it (with the compiled regexes made global) for now.
    content = TITLE_TAG_DATA.sub(r'<\1title>', content)
    content = QUOTED_TITLE.sub('', content)

    start = content.rfind('<title>')
    end = content.rfind('</title>')
    if start == -1 or end == -1:
        return
    title = web.decode(content[start + 7:end])
    title = title.strip()[:200]

    title = ' '.join(title.split())  # cleanly remove multiple spaces

    # More cryptic regex substitutions. This one looks to be myano's invention.
    title = RE_DCC.sub('', title)

    return title or None


def get_hostname(url):
    idx = 7
    if url.startswith('https://'):
        idx = 8
    elif url.startswith('ftp://'):
        idx = 6
    hostname = url[idx:]
    slash = hostname.find('/')
    if slash != -1:
        hostname = hostname[:slash]
    return hostname


def get_or_create_shorturl(bot, url):
    """Get or create a short URL for ``url``

    :param bot: Sopel instance
    :param str url: URL to get or create a short URL for
    :return: A short URL
    :rtype: str

    It gets the short URL for ``url`` from the bot's memory if it exists.
    Otherwise, it creates a short URL (see :func:`get_tinyurl`), stores it
    into the bot's memory, then returns it.
    """
    # Check bot memory to see if the shortened URL is already in
    # memory
    if url in bot.memory['shortened_urls']:
        return bot.memory['shortened_urls'][url]

    tinyurl = get_tinyurl(url)
    bot.memory['shortened_urls'][url] = tinyurl
    return tinyurl


def get_tinyurl(url):
    """Returns a shortened tinyURL link of the URL"""
    base_url = "https://tinyurl.com/api-create.php"
    tinyurl = "%s?%s" % (base_url, web.urlencode({'url': url}))
    try:
        res = requests.get(tinyurl)
        res.raise_for_status()
    except requests.exceptions.RequestException:
        return None
    # Replace text output with https instead of http to make the
    # result an HTTPS link.
    return res.text.replace("http://", "https://")



# coding=utf-8
"""
find.py - Sopel Spelling Correction Plugin
This plugin will fix spelling errors if someone corrects them
using the sed notation (s///) commonly found in vi/vim.

Copyright 2011, Michael Yanovich, yanovich.net
Copyright 2013, Elsie Powell, embolalia.com
Copyright 2020, dgw, technobabbl.es
Includes contributions from: Matt Meinwald, and Morgan Goose
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import deque
import re

from sopel import plugin
from sopel.formatting import bold
from sopel.tools import Identifier, SopelIdentifierMemory


def setup(bot):
    if 'find_lines' not in bot.memory:
        bot.memory['find_lines'] = SopelIdentifierMemory()


def shutdown(bot):
    try:
        del bot.memory['find_lines']
    except KeyError:
        pass


@plugin.echo
@plugin.rule('.*')
@plugin.priority('low')
@plugin.require_chanmsg
@plugin.unblockable
def collectlines(bot, trigger):
    """Create a temporary log of what people say"""
    # Add a log for the channel and nick, if there isn't already one
    if trigger.sender not in bot.memory['find_lines']:
        bot.memory['find_lines'][trigger.sender] = SopelIdentifierMemory()
    if trigger.nick not in bot.memory['find_lines'][trigger.sender]:
        bot.memory['find_lines'][trigger.sender][trigger.nick] = deque(maxlen=10)

    # Update in-memory list of the user's lines in the channel
    line_list = bot.memory['find_lines'][trigger.sender][trigger.nick]
    line = trigger.group()
    if line.startswith('s/') or line.startswith('s|'):
        # Don't remember substitutions
        return
    # store messages in reverse order (most recent first)
    elif line.startswith('\x01ACTION'):  # For /me messages
        line = line[:-1]
        line_list.appendleft(line)
    else:
        line_list.appendleft(line)


def _cleanup_channel(bot, channel):
    bot.memory['find_lines'].pop(channel, None)


def _cleanup_nickname(bot, nick, channel=None):
    if channel:
        bot.memory['find_lines'].get(channel, {}).pop(nick, None)
    else:
        for channel in bot.memory['find_lines'].keys():
            bot.memory['find_lines'][channel].pop(nick, None)


@plugin.echo
@plugin.event('PART')
@plugin.priority('low')
@plugin.unblockable
def part_cleanup(bot, trigger):
    """Clean up cached data when a user leaves a channel."""
    if trigger.nick == bot.nick:
        # Nuke the whole channel cache, boys, we're outta here!
        _cleanup_channel(bot, trigger.sender)
    else:
        # Someone else left; clean up after them
        _cleanup_nickname(bot, trigger.nick, trigger.sender)


@plugin.echo
@plugin.event('QUIT')
@plugin.priority('low')
@plugin.unblockable
def quit_cleanup(bot, trigger):
    """Clean up cached data after a user quits IRC."""
    # If Sopel itself quits, shutdown() will handle the cleanup.
    _cleanup_nickname(bot, trigger.nick)


@plugin.echo
@plugin.event('KICK')
@plugin.priority('low')
@plugin.unblockable
def kick_cleanup(bot, trigger):
    """Clean up cached data when a user is kicked from a channel."""
    nick = Identifier(trigger.args[1])
    if nick == bot.nick:
        # We got kicked! Nuke the whole channel.
        _cleanup_channel(bot, trigger.sender)
    else:
        # Clean up after the poor sod (or more likely, spammer) who got the boot
        _cleanup_nickname(bot, nick, trigger.sender)


# Match nick, s/find/replace/flags. Flags and nick are optional, nick can be
# followed by comma or colon, anything after the first space after the third
# slash is ignored, and you can use either a slash or a pipe.
# If you want to search for an actual slash AND a pipe in the same message,
# you can escape your separator, in old and/or new.
@plugin.rule(r"""(?:
             (?P<nick>\S+)    # Catch a nick in group 1
             [:,]\s+)?        # Followed by optional colon/comma and whitespace
             s(?P<sep>/)      # The literal s and a separator / as group 2
             (?P<old>         # Group 3 is the thing to find
               (?:\\/|[^/])+  # One or more non-slashes or escaped slashes
             )
             /                # The separator again
             (?P<new>         # Group 4 is what to replace with
               (?:\\/|[^/])*  # One or more non-slashes or escaped slashes
             )
             (?:/             # Optional separator followed by group 5 (flags)
                (?P<flags>\S+)
             )?
            """)
@plugin.rule(r"""(?:
             (?P<nick>\S+)    # Catch a nick in group 1
             [:,]\s+)?        # Followed by optional colon/comma and whitespace
             s(?P<sep>\|)     # The literal s and a separator | as group 2
             (?P<old>         # Group 3 is the thing to find
               (?:\\\||[^|])+  # One or more non-pipe or escaped pipe
             )
             \|               # The separator again
             (?P<new>         # Group 4 is what to replace with
               (?:\\\||[^|])* # One or more non-pipe or escaped pipe
             )
             (?:|             # Optional separator followed by group 5 (flags)
                (?P<flags>\S+)
             )?
            """)
@plugin.priority('high')
def findandreplace(bot, trigger):
    # Don't bother in PM
    if trigger.is_privmsg:
        return

    # Correcting other person vs self.
    rnick = Identifier(trigger.group('nick') or trigger.nick)

    # only do something if there is conversation to work with
    history = bot.memory['find_lines'].get(trigger.sender, {}).get(rnick, None)
    if not history:
        return

    sep = trigger.group('sep')
    old = trigger.group('old').replace('\\%s' % sep, sep)
    new = trigger.group('new')
    me = False  # /me command
    flags = trigger.group('flags') or ''

    # only clean/format the new string if it's non-empty
    if new:
        new = bold(new.replace('\\%s' % sep, sep))

    # If g flag is given, replace all. Otherwise, replace once.
    if 'g' in flags:
        count = -1
    else:
        count = 1

    # repl is a dynamically defined function which performs the substitution.
    # i flag turns off case sensitivity. re.U turns on unicode replacement.
    if 'i' in flags:
        regex = re.compile(re.escape(old), re.U | re.I)

        def repl(s):
            return re.sub(regex, new, s, count == 1)
    else:
        def repl(s):
            return s.replace(old, new, count)

    # Look back through the user's lines in the channel until you find a line
    # where the replacement works
    new_phrase = None
    for line in history:
        if line.startswith("\x01ACTION"):
            me = True  # /me command
            line = line[8:]
        else:
            me = False
        replaced = repl(line)
        if replaced != line:  # we are done
            new_phrase = replaced
            break

    if not new_phrase:
        return  # Didn't find anything

    # Save the new "edited" message.
    action = (me and '\x01ACTION ') or ''  # If /me message, prepend \x01ACTION
    history.appendleft(action + new_phrase)  # history is in most-recent-first order

    # output
    if not me:
        new_phrase = 'meant to say: %s' % new_phrase
    if trigger.group(1):
        phrase = '%s thinks %s %s' % (trigger.nick, rnick, new_phrase)
    else:
        phrase = '%s %s' % (trigger.nick, new_phrase)

    bot.say(phrase)



#!/usr/bin/env python
# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import sys

try:
    from setuptools import setup, __version__ as setuptools_version
except ImportError:
    print(
        'You do not have setuptools, and can not install Sopel. The easiest '
        'way to fix this is to install pip by following the instructions at '
        'https://pip.readthedocs.io/en/latest/installing/\n'
        'Alternately, you can run Sopel without installing it by running '
        '"python sopel.py"',
        file=sys.stderr,
    )
    sys.exit(1)
else:
    version_info = setuptools_version.split('.')
    major = int(version_info[0])
    minor = int(version_info[1])

    if major < 30 or (major == 30 and minor < 3):
        print(
            'Your version of setuptools is outdated: version 30.3 or above '
            'is required to install Sopel. You can do that with '
            '"pip install -U setuptools"\n'
            'Alternately, you can run Sopel without installing it by running '
            '"python sopel.py"',
            file=sys.stderr,
        )
        sys.exit(1)

# We check Python's version ourselves in case someone installed Sopel on an
# old version of pip (<9.0.0), which doesn't know about `python_requires`.
if sys.version_info < (2, 7) or (
    sys.version_info.major >= 3 and sys.version_info < (3, 3)
):
    # Maybe not the best way to do this, but this question is tiring.
    raise ImportError('Sopel requires Python 2.7+ or 3.3+.')
# Py2 EOL: https://www.python.org/dev/peps/pep-0373/#maintenance-releases
if sys.version_info.major == 2:
    print(
        'Warning: Python 2.x has reached end of life and will receive '
        'no further updates. Sopel 8.0 will drop support for it.',
        file=sys.stderr,
    )


def read_reqs(path):
    with open(path, 'r') as fil:
        return list(fil.readlines())


requires = read_reqs('requirements.txt')
if sys.version_info[0] < 3:
    requires.append('backports.ssl_match_hostname')
dev_requires = requires + read_reqs('dev-requirements.txt')

setup(
    install_requires=requires, extras_require={"dev": dev_requires},
)



# coding=utf-8
"""Sopel's Job Scheduler: internal tool for job management.

.. important::

    As of Sopel 5.3, this is an internal tool used by Sopel to manage internal
    jobs and should not be used by plugin authors. Its usage and documentation
    is for Sopel core development and advanced developers. It is subject to
    rapid changes between versions without much (or any) warning.

"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import logging
import threading
import time


LOGGER = logging.getLogger(__name__)


class Scheduler(threading.Thread):
    """Generic Job Scheduler.

    :param object manager: manager passed to jobs as argument

    Scheduler is a :class:`thread <threading.Thread>` that keeps track of
    :class:`Jobs <Job>` and periodically checks which ones are ready to
    execute. When ready, their :meth:`~Job.execute` method is called, either
    in a separate thread or in the scheduler's thread (it depends on the
    job's :meth:`~Job.is_threaded` method).

    It can be started as any other thread::

        # on bot's startup
        scheduler = jobs.Scheduler(bot)
        scheduler.start()  # run the thread forever

    Then it runs forever until the :meth:`stop` method is called, usually when
    the bot shuts down.

    .. note::

        Thread safety is ensured with threading's :class:`~threading.Lock`
        and :class:`~threading.Event` when:

        * a job is :meth:`registered <register>` or
          :meth:`removed <remove_callable_job>`
        * the scheduler is :meth:`cleared <clear_jobs>` or
          :meth:`stopped <stop>`
        * the scheduler gets jobs that are ready for execution

        These actions can be performed while the scheduler is running.

    .. important::

        This is an internal tool used by Sopel to manage internal jobs and
        should not be used by plugin authors. Its usage and documentation is
        for Sopel core development and advanced developers. It is subject to
        rapid changes between versions without much (or any) warning.

    """
    def __init__(self, manager):
        threading.Thread.__init__(self)
        self.manager = manager
        """Job manager, used as argument for jobs."""
        self.stopping = threading.Event()
        """Stopping flag. See :meth:`stop`."""
        self._jobs = []
        self._mutex = threading.Lock()

    def register(self, job):
        """Register a Job to the current job queue.

        :param job: job to register
        :type job: :class:`sopel.tools.jobs.Job`

        This method is thread safe.
        """
        with self._mutex:
            self._jobs.append(job)
        LOGGER.debug('Job registered: %s', str(job))

    def clear_jobs(self):
        """Clear current Job queue and start fresh.

        This method is thread safe. However, it won't cancel or stop any
        currently running jobs.
        """
        with self._mutex:
            self._jobs = []

    def stop(self):
        """Ask the job scheduler to stop.

        The scheduler thread will stop its loop over jobs to process, but it
        won't join the thread, or clear its queue—this has to be done
        separately by the calling thread::

            scheduler.stop()  # ask the scheduler to stop
            scheduler.join()  # wait for the scheduler to actually stop

        Note that this won't cancel or stop any currently running jobs.
        """
        self.stopping.set()

    def remove_callable_job(self, callable):
        """Remove ``callable`` from the job queue.

        :param callable callable: the callable to remove
        :type callable: :term:`function`

        This method is thread safe. However, it won't cancel or stop any
        currently running jobs.
        """
        with self._mutex:
            self._jobs = [
                job for job in self._jobs
                if job._handler != callable
            ]

    def run(self):
        """Run forever until :meth:`stop` is called.

        This method waits at most a second between each iteration. At each step
        it retrieves the jobs that are ready for execution, and executes them.
        See the :meth:`Job.execute` method for more information.

        Internally, it loops forever until its :attr:`stopping` event is set.

        .. note::

            This should not be called directly, as it will be done by the
            :meth:`threading.Thread.start` method.

        """
        while not self.stopping.is_set():
            try:
                now = time.time()

                # Collect ready jobs by now
                for job in self._get_ready_jobs(now):
                    self._run_job(job)

                # Wait up to a second
                time_spent = time.time() - now
                wait_time = max(0, 1 - time_spent)
                if wait_time:
                    time.sleep(wait_time)
            except KeyboardInterrupt:
                # Do not block on KeyboardInterrupt
                LOGGER.debug('Job scheduler stopped by KeyboardInterrupt')
                raise
            except Exception as error:  # TODO: Be specific
                LOGGER.error('Error in job scheduler: %s', error)
                # Plugins exceptions are caught earlier, so this is a bit
                # more serious. Options are to either stop the main thread
                # or continue this thread and hope that it won't happen
                # again.
                self.manager.on_scheduler_error(self, error)
                # Sleep a bit to guard against busy-looping and filling
                # the log with useless error messages.
                time.sleep(10.0)  # seconds

    def _get_ready_jobs(self, now):
        with self._mutex:
            jobs = [job for job in self._jobs if job.is_ready_to_run(now)]

        return jobs

    def _run_job(self, job):
        if job.is_threaded():
            # make sure the job knows it's running, even though the thread
            # isn't started yet.
            job.is_running.set()
            t = threading.Thread(
                target=self._call, args=(job,)
            )
            t.start()
        else:
            self._call(job)

    def _call(self, job):
        """Wrap the job's execution to handle its state and errors."""
        try:
            with job:
                job.execute(self.manager)
        except Exception as error:  # TODO: Be specific
            LOGGER.error('Error while processing job: %s', error)
            self.manager.on_job_error(self, job, error)


class Job(object):
    """Holds information about when a function should be called next.

    :param intervals: set of intervals; each is a number of seconds between
                      calls to ``handler``
    :type intervals: :term:`iterable`
    :param str plugin: optional plugin name to which the job belongs
    :param str label: optional label (name) for the job
    :param handler: function to be called when the job is ready to execute
    :type handler: :term:`function`
    :param str doc: optional documentation for the job

    Job is a simple structure that holds information about when a function
    should be called next. They are best used with a :class:`Scheduler`
    that will manage job execution when they are ready.

    The :term:`function` to execute is the ``handler``, which must be a
    callable with this signature::

        def handler(manager):
            # perform action periodically
            # return is optional

    The ``manager`` parameter can be any kind of object; usually it's an
    instance of :class:`sopel.bot.Sopel`.

    When a job is ready, you can execute it by calling its :meth:`execute`
    method (providing the appropriate ``manager`` argument)::

        if job.is_ready_to_run(time.time()):
            job.execute(manager)  # marked as running
            # "next times" have been updated; the job is not running

    In that case, ``execute`` takes care of the running state of the job.

    Alternatively, you can use a ``with`` statement to perform action before
    and/or after executing the job; in that case, the ``with`` statement takes
    precedence, and the :meth:`execute` method won't interfere::

        with job:
            # the job is now running, you can perform pre-execute action
            job.execute()  # execute the job's action, no state modification
            # the job is still marked as "running"
            # you can perform post-execute action

        # outside of the with statement, the job is not running anymore

    .. seealso::

        The :class:`sopel.plugins.jobs.Scheduler` class is specifically
        designed for plugins' jobs, expecting an instance of
        :class:`sopel.bot.Sopel` as a manager, and should be used to manipulate
        plugin jobs.

        In all other case, the :class:`sopel.tools.jobs.Scheduler` class is a
        generic job scheduler.

    """
    @classmethod
    def kwargs_from_callable(cls, handler):
        """Generate the keyword arguments to create a new instance.

        :param handler: callable used to generate keyword arguments
        :type handler: :term:`function`
        :return: a map of keyword arguments
        :rtype: dict

        This classmethod takes the ``handler``'s attributes to generate a map
        of keyword arguments for the class. This can be used by the
        :meth:`from_callable` classmethod to instantiate a new rule object.

        The expected attributes are the ones set by decorators from the
        :mod:`sopel.plugin` module.
        """

        return {
            'plugin': getattr(handler, 'plugin_name', None),
            'label': getattr(handler, 'rule_label', None),
            'threaded': getattr(handler, 'thread', True),
            'doc': inspect.getdoc(handler),
        }

    @classmethod
    def from_callable(cls, settings, handler):
        """Instantiate a Job from the bot's ``settings`` and a ``handler``.

        :param settings: bot's settings
        :type settings: :class:`sopel.config.Config`
        :param handler: callable used to instantiate a new job
        :type handler: :term:`function`
        """
        kwargs = cls.kwargs_from_callable(handler)
        return cls(
            set(handler.interval),
            handler=handler,
            **kwargs)

    def __init__(self,
                 intervals,
                 plugin=None,
                 label=None,
                 handler=None,
                 threaded=True,
                 doc=None):
        # scheduling
        now = time.time()
        self.intervals = set(intervals)
        """Set of intervals at which to execute the job."""
        self.next_times = dict(
            (interval, now + interval)
            for interval in self.intervals
        )
        """Tracking of when to execute the job next time."""

        # meta
        self._plugin_name = plugin
        self._label = label
        self._doc = doc

        # execution
        self._handler = handler
        self._threaded = bool(threaded)
        self.is_running = threading.Event()
        """Running flag: it tells if the job is running or not.

        This flag is set and cleared automatically by the :meth:`execute`
        method. It is also set and cleared when the job is used with the
        ``with`` statement::

            with job:
                # you do something before executing the job
                # this ensures that the job is marked as "running"

        .. note::

            When set manually or with the ``with`` statement, the
            :meth:`execute` method won't clear this attribute itself.

        """

    def __enter__(self):
        self.is_running.set()

    def __exit__(self, exc_type, exc_value, traceback):
        self.next(time.time())
        self.is_running.clear()

    def __str__(self):
        """Return a string representation of the Job object.

        Example result::

            <Job periodic_check [5s]>
            <Job periodic_check [60s, 3600s]>

        Example when the job is tied to a plugin::

            <Job reminder.remind_check [2s]>

        """
        try:
            label = self.get_job_label()
        except RuntimeError:
            label = '(unknown)'

        plugin_name = self.get_plugin_name()
        if plugin_name:
            label = '%s.%s' % (plugin_name, label)

        return "<Job %s [%s]>" % (
            label,
            ', '.join('%ss' % i for i in sorted(self.intervals)),
        )

    def get_plugin_name(self):
        """Get the job's plugin name.

        :rtype: str

        The job's plugin name will be used in various places to select,
        register, unregister, and manipulate the job based on its plugin, which
        is referenced by its name.
        """
        return self._plugin_name

    def get_job_label(self):
        """Get the job's label.

        :rtype: str

        A job can have a label, which can identify the job by string, the same
        way rules can be. This label can be used to manipulate or display the
        job's information in a more human-readable way. Note that the label has
        no effect on the job's execution.
        """
        if self._label:
            return self._label

        if self._handler is not None and self._handler.__name__:
            return self._handler.__name__

        raise RuntimeError('Undefined job label')

    def get_doc(self):
        """Get the job's documentation.

        :rtype: str

        A job's documentation is a short text that can be displayed to a user.
        """
        return self._doc

    def is_threaded(self):
        """Tell if the job's execution should be in a thread.

        :return: ``True`` if the execution should be in a thread,
                 ``False`` otherwise
        :rtype: bool
        """
        return self._threaded

    def is_ready_to_run(self, at_time):
        """Check if this job is (or will be) ready to run at the given time.

        :param int at_time: Timestamp to check, in seconds
        :return: ``True`` if the job is (or will be) ready to run, ``False``
                 otherwise
        :rtype: bool
        """
        return not self.is_running.is_set() and any(
            (next_time - at_time) <= 0
            for next_time in self.next_times.values()
        )

    def next(self, current_time):
        """Update :attr:`next_times`, assuming it executed at ``current_time``.

        :param int current_time: timestamp of the current time
        :return: a modified job object
        """
        for interval, last_time in list(self.next_times.items()):
            if last_time >= current_time:
                # no need to update this interval
                continue

            # if last time + interval is in the future, it's used
            # else, try to run it asap
            self.next_times[interval] = max(last_time + interval, current_time)

        return self

    def execute(self, manager):
        """Execute the job's handler and return its result.

        :param object manager: used as argument to the job's handler
        :return: the return value from the handler's execution

        This method executes the job's handler. It doesn't change its running
        state, as this must be done by the caller::

            with job:  # mark as running
                # before execution
                job.execute(manager)
                # after execution

        """
        return self._handler(manager)



# coding=utf-8
"""Sopel's plugins interface.

.. versionadded:: 7.0

Sopel uses what are called Plugin Handlers as an interface between the bot and
its plugins (formerly called "modules"). This interface is defined by the
:class:`~.handlers.AbstractPluginHandler` abstract class.

Plugins that can be used by Sopel are provided by :func:`~.get_usable_plugins`
in an :class:`ordered dict<collections.OrderedDict>`. This dict contains one
and only one plugin per unique name, using a specific order:

* extra directories defined in the settings
* homedir's ``plugins`` directory
* homedir's ``modules`` directory
* ``sopel.plugins`` setuptools entry points
* ``sopel_modules``'s subpackages
* ``sopel.modules``'s core plugins

(The ``coretasks`` plugin is *always* the one from ``sopel.coretasks`` and
cannot be overridden.)

To find all plugins (no matter their sources), the :func:`~.enumerate_plugins`
function can be used. For a more fine-grained search, ``find_*`` functions
exist for each type of plugin.
"""
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import imp
import itertools
import os

import pkg_resources

from . import exceptions, handlers, rules  # noqa


def _list_plugin_filenames(directory):
    # list plugin filenames from a directory
    # yield 2-value tuples: (name, absolute path)
    base = os.path.abspath(directory)
    for filename in os.listdir(base):
        abspath = os.path.join(base, filename)

        if os.path.isdir(abspath):
            if os.path.isfile(os.path.join(abspath, '__init__.py')):
                yield os.path.basename(filename), abspath
        else:
            name, ext = os.path.splitext(filename)
            if ext == '.py' and name != '__init__':
                yield name, abspath


def find_internal_plugins():
    """List internal plugins.

    :return: yield instances of :class:`~.handlers.PyModulePlugin`
             configured for ``sopel.modules.*``

    Internal plugins can be found under ``sopel.modules``. This list does not
    include the ``coretasks`` plugin.
    """
    plugin_dir = imp.find_module(
        'modules',
        [imp.find_module('sopel')[1]]
    )[1]

    for name, _ in _list_plugin_filenames(plugin_dir):
        yield handlers.PyModulePlugin(name, 'sopel.modules')


def find_sopel_modules_plugins():
    """List plugins from ``sopel_modules.*``.

    :return: yield instances of :class:`~.handlers.PyModulePlugin`
             configured for ``sopel_modules.*``

    Before entry point plugins, the only way to package a plugin was to follow
    :pep:`382` by using the ``sopel_modules`` namespace. This function is
    responsible to load such plugins.
    """
    try:
        import sopel_modules
    except ImportError:
        return

    for plugin_dir in set(sopel_modules.__path__):
        for name, _ in _list_plugin_filenames(plugin_dir):
            yield handlers.PyModulePlugin(name, 'sopel_modules')


def find_entry_point_plugins(group='sopel.plugins'):
    """List plugins from a setuptools entry point group.

    :param str group: setuptools entry point group to look for
                      (defaults to ``sopel.plugins``)
    :return: yield instances of :class:`~.handlers.EntryPointPlugin`
             created from setuptools entry point given ``group``

    This function finds plugins declared under a setuptools entry point; by
    default it uses the ``sopel.plugins`` entry point.
    """
    for entry_point in pkg_resources.iter_entry_points(group):
        yield handlers.EntryPointPlugin(entry_point)


def find_directory_plugins(directory):
    """List plugins from a ``directory``.

    :param str directory: directory path to search
    :return: yield instances of :class:`~.handlers.PyFilePlugin`
             found in ``directory``

    This function looks for single file and folder plugins in a directory.
    """
    for _, abspath in _list_plugin_filenames(directory):
        yield handlers.PyFilePlugin(abspath)


def enumerate_plugins(settings):
    """Yield Sopel's plugins.

    :param settings: Sopel's configuration
    :type settings: :class:`sopel.config.Config`
    :return: yield 2-value tuple: an instance of
             :class:`~.handlers.AbstractPluginHandler`, and if the plugin is
             active or not

    This function uses the find functions to find all of Sopel's available
    plugins. It uses the bot's ``settings`` to determine if the plugin is
    enabled or disabled.

    .. seealso::

       The find functions used are:

       * :func:`find_internal_plugins` for internal plugins
       * :func:`find_sopel_modules_plugins` for ``sopel_modules.*`` plugins
       * :func:`find_entry_point_plugins` for plugins exposed by setuptools
         entry points
       * :func:`find_directory_plugins` for plugins in ``$homedir/modules``,
         ``$homedir/plugins``, and in extra directories, as defined by
         ``settings.core.extra``

    .. versionchanged:: 7.0

       Previously, plugins were called "modules", so this would load plugins
       from the ``$homedir/modules`` directory. Now it also loads plugins
       from the ``$homedir/plugins`` directory.

    """
    from_internals = find_internal_plugins()
    from_sopel_modules = find_sopel_modules_plugins()
    from_entry_points = find_entry_point_plugins()
    # load from directories
    source_dirs = [
        os.path.join(settings.homedir, 'modules'),
        os.path.join(settings.homedir, 'plugins'),
    ]
    if settings.core.extra:
        source_dirs = source_dirs + list(settings.core.extra)

    from_directories = [
        find_directory_plugins(source_dir)
        for source_dir in source_dirs
        if os.path.isdir(source_dir)
    ]

    # Retrieve all plugins
    all_plugins = itertools.chain(
        from_internals,
        from_sopel_modules,
        from_entry_points,
        *from_directories)

    # Get plugin settings
    enabled = settings.core.enable
    disabled = settings.core.exclude

    # Yield all found plugins with their enabled status (True/False)
    for plugin in all_plugins:
        name = plugin.name
        is_enabled = name not in disabled and (not enabled or name in enabled)
        yield plugin, is_enabled

    # And always yield coretasks
    yield handlers.PyModulePlugin('coretasks', 'sopel'), True


def get_usable_plugins(settings):
    """Get usable plugins, unique per name.

    :param settings: Sopel's configuration
    :type settings: :class:`sopel.config.Config`
    :return: an ordered dict of usable plugins
    :rtype: collections.OrderedDict

    This function provides the plugins Sopel can use to load, enable,
    or disable, as an :class:`ordered dict<collections.OrderedDict>`. This dict
    contains one and only one plugin per unique name, using a specific order:

    * extra directories defined in the settings
    * homedir's ``modules`` directory
    * ``sopel.plugins`` setuptools entry points
    * ``sopel_modules``'s subpackages
    * ``sopel.modules``'s core plugins

    (The ``coretasks`` plugin is *always* the one from ``sopel.coretasks`` and
    cannot be overridden.)

    .. seealso::

        The :func:`~.enumerate_plugins` function is used to generate a list
        of all possible plugins, and its return value is used to populate
        the :class:`ordered dict<collections.OrderedDict>`.

    """
    # Use an OrderedDict to get one and only one plugin per name
    # based on what plugins.enumerate_plugins does, external plugins are
    # allowed to override internal plugins
    plugins_info = collections.OrderedDict(
        (plugin.name, (plugin, is_enabled))
        for plugin, is_enabled in enumerate_plugins(settings))
    # reset coretasks's position at the end of the loading queue
    # Python 2's OrderedDict does not have a `move_to_end` method
    # TODO: replace by plugins_info.move_to_end('coretasks') for Python 3
    core_info = plugins_info.pop('coretasks')
    plugins_info['coretasks'] = core_info

    return plugins_info



# coding=utf-8
# Copyright 2008, Sean B. Palmer, inamidst.com
# Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
# Copyright 2012-2015, Elsie Powell, http://embolalia.com
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.

from __future__ import absolute_import, division, print_function, unicode_literals

from ast import literal_eval
from datetime import datetime
import inspect
import itertools
import logging
import re
import signal
import sys
import threading
import time

from sopel import irc, logger, plugins, tools
from sopel.db import SopelDB
import sopel.loader
from sopel.module import NOLIMIT
from sopel.plugins import jobs as plugin_jobs, rules as plugin_rules
from sopel.tools import deprecated, Identifier
import sopel.tools.jobs
from sopel.trigger import Trigger


__all__ = ['Sopel', 'SopelWrapper']

LOGGER = logging.getLogger(__name__)
QUIT_SIGNALS = [
    getattr(signal, name)
    for name in ['SIGUSR1', 'SIGTERM', 'SIGINT']
    if hasattr(signal, name)
]
RESTART_SIGNALS = [
    getattr(signal, name)
    for name in ['SIGUSR2', 'SIGILL']
    if hasattr(signal, name)
]
SIGNALS = QUIT_SIGNALS + RESTART_SIGNALS


if sys.version_info.major >= 3:
    unicode = str
    basestring = str
    py3 = True
else:
    py3 = False


class Sopel(irc.AbstractBot):
    def __init__(self, config, daemon=False):
        super(Sopel, self).__init__(config)
        self._daemon = daemon  # Used for iPython. TODO something saner here
        self.wantsrestart = False
        self._running_triggers = []
        self._running_triggers_lock = threading.Lock()
        self._plugins = {}
        self._rules_manager = plugin_rules.Manager()
        self._scheduler = plugin_jobs.Scheduler(self)

        self._times = {}
        """
        A dictionary mapping lowercased nicks to dictionaries which map
        function names to the time which they were last used by that nick.
        """

        self.server_capabilities = {}
        """A dict mapping supported IRCv3 capabilities to their options.

        For example, if the server specifies the capability ``sasl=EXTERNAL``,
        it will be here as ``{"sasl": "EXTERNAL"}``. Capabilities specified
        without any options will have ``None`` as the value.

        For servers that do not support IRCv3, this will be an empty set.
        """

        self.privileges = dict()
        """A dictionary of channels to their users and privilege levels.

        The value associated with each channel is a dictionary of
        :class:`sopel.tools.Identifier`\\s to a bitwise integer value,
        determined by combining the appropriate constants from
        :mod:`sopel.plugin`.

        .. deprecated:: 6.2.0
            Use :attr:`channels` instead. Will be removed in Sopel 8.
        """

        self.channels = tools.SopelIdentifierMemory()
        """A map of the channels that Sopel is in.

        The keys are :class:`sopel.tools.Identifier`\\s of the channel names,
        and map to :class:`sopel.tools.target.Channel` objects which contain
        the users in the channel and their permissions.
        """

        self.users = tools.SopelIdentifierMemory()
        """A map of the users that Sopel is aware of.

        The keys are :class:`sopel.tools.Identifier`\\s of the nicknames, and
        map to :class:`sopel.tools.target.User` instances. In order for Sopel
        to be aware of a user, it must share at least one mutual channel.
        """

        self.db = SopelDB(config)
        """The bot's database, as a :class:`sopel.db.SopelDB` instance."""

        self.memory = tools.SopelMemory()
        """
        A thread-safe dict for storage of runtime data to be shared between
        plugins. See :class:`sopel.tools.SopelMemory`.
        """

        self.shutdown_methods = []
        """List of methods to call on shutdown."""

    @property
    def rules(self):
        """Rules manager."""
        return self._rules_manager

    @property
    def scheduler(self):
        """Job Scheduler. See :func:`sopel.plugin.interval`."""
        return self._scheduler

    @property
    def command_groups(self):
        """A mapping of plugin names to lists of their commands.

        .. versionchanged:: 7.1
            This attribute is now generated on the fly from the registered list
            of commands and nickname commands.
        """
        # This was supposed to be deprecated, but the built-in help plugin needs it
        # TODO: create a new, better, doc interface to remove it
        plugin_commands = itertools.chain(
            self._rules_manager.get_all_commands(),
            self._rules_manager.get_all_nick_commands(),
        )
        result = {}

        for plugin, commands in plugin_commands:
            if plugin not in result:
                result[plugin] = list(sorted(commands.keys()))
            else:
                result[plugin].extend(commands.keys())
                result[plugin] = list(sorted(result[plugin]))

        return result

    @property
    def doc(self):
        """A dictionary of command names to their documentation.

        Each command is mapped to its docstring and any available examples, if
        declared in the plugin's code.

        .. versionchanged:: 3.2
            Use the first item in each callable's commands list as the key,
            instead of the function name as declared in the source code.

        .. versionchanged:: 7.1
            This attribute is now generated on the fly from the registered list
            of commands and nickname commands.
        """
        # TODO: create a new, better, doc interface to remove it
        plugin_commands = itertools.chain(
            self._rules_manager.get_all_commands(),
            self._rules_manager.get_all_nick_commands(),
        )
        commands = (
            (command, command.get_doc(), command.get_usages())
            for plugin, commands in plugin_commands
            for command in commands.values()
        )

        return dict(
            (name, (doc.splitlines(), [u['text'] for u in usages]))
            for command, doc, usages in commands
            for name in ((command.name,) + command.aliases)
        )

    @property
    def hostmask(self):
        """The current hostmask for the bot :class:`sopel.tools.target.User`.

        :return: the bot's current hostmask
        :rtype: str

        Bot must be connected and in at least one channel.
        """
        if not self.users or self.nick not in self.users:
            raise KeyError("'hostmask' not available: bot must be connected and in at least one channel.")

        return self.users.get(self.nick).hostmask

    def has_channel_privilege(self, channel, privilege):
        """Tell if the bot has a ``privilege`` level or above in a ``channel``.

        :param str channel: a channel the bot is in
        :param int privilege: privilege level to check
        :raise ValueError: when the channel is unknown

        This method checks the bot's privilege level in a channel, i.e. if it
        has this level or higher privileges::

            >>> bot.channels['#chan'].privileges[bot.nick] = plugin.OP
            >>> bot.has_channel_privilege('#chan', plugin.VOICE)
            True

        The ``channel`` argument can be either a :class:`str` or a
        :class:`sopel.tools.Identifier`, as long as Sopel joined said channel.
        If the channel is unknown, a :exc:`ValueError` will be raised.
        """
        if channel not in self.channels:
            raise ValueError('Unknown channel %s' % channel)

        return self.channels[channel].has_privilege(self.nick, privilege)

    # signal handlers

    def set_signal_handlers(self):
        """Set signal handlers for the bot.

        Before running the bot, this method can be called from the main thread
        to setup signals. If the bot is connected, upon receiving a signal it
        will send a ``QUIT`` message. Otherwise, it raises a
        :exc:`KeyboardInterrupt` error.

        .. note::

            Per the Python documentation of :func:`signal.signal`:

                When threads are enabled, this function can only be called from
                the main thread; attempting to call it from other threads will
                cause a :exc:`ValueError` exception to be raised.

        """
        for obj in SIGNALS:
            signal.signal(obj, self._signal_handler)

    def _signal_handler(self, sig, frame):
        if sig in QUIT_SIGNALS:
            if self.backend.is_connected():
                LOGGER.warning("Got quit signal, sending QUIT to server.")
                self.quit('Closing')
            else:
                self.hasquit = True  # mark the bot as "want to quit"
                LOGGER.warning("Got quit signal.")
                raise KeyboardInterrupt
        elif sig in RESTART_SIGNALS:
            if self.backend.is_connected():
                LOGGER.warning("Got restart signal, sending QUIT to server.")
                self.restart('Restarting')
            else:
                LOGGER.warning("Got restart signal.")
                self.wantsrestart = True  # mark the bot as "want to restart"
                self.hasquit = True  # mark the bot as "want to quit"
                raise KeyboardInterrupt

    # setup

    def setup(self):
        """Set up Sopel bot before it can run.

        The setup phase is in charge of:

        * setting up logging (configure Python's built-in :mod:`logging`)
        * setting up the bot's plugins (load, setup, and register)
        * starting the job scheduler
        """
        self.setup_logging()
        self.setup_plugins()
        self.post_setup()

    def setup_logging(self):
        """Set up logging based on config options."""
        logger.setup_logging(self.settings)
        base_level = self.settings.core.logging_level or 'INFO'
        base_format = self.settings.core.logging_format
        base_datefmt = self.settings.core.logging_datefmt

        # configure channel logging if required by configuration
        if self.settings.core.logging_channel:
            channel_level = self.settings.core.logging_channel_level or base_level
            channel_format = self.settings.core.logging_channel_format or base_format
            channel_datefmt = self.settings.core.logging_channel_datefmt or base_datefmt
            channel_params = {}
            if channel_format:
                channel_params['fmt'] = channel_format
            if channel_datefmt:
                channel_params['datefmt'] = channel_datefmt
            formatter = logger.ChannelOutputFormatter(**channel_params)
            handler = logger.IrcLoggingHandler(self, channel_level)
            handler.setFormatter(formatter)

            # set channel handler to `sopel` logger
            LOGGER = logging.getLogger('sopel')
            LOGGER.addHandler(handler)

    def setup_plugins(self):
        """Load plugins into the bot."""
        load_success = 0
        load_error = 0
        load_disabled = 0

        LOGGER.info("Loading plugins...")
        usable_plugins = plugins.get_usable_plugins(self.settings)
        for name, info in usable_plugins.items():
            plugin, is_enabled = info
            if not is_enabled:
                load_disabled = load_disabled + 1
                continue

            try:
                plugin.load()
            except Exception as e:
                load_error = load_error + 1
                LOGGER.exception("Error loading %s: %s", name, e)
            except SystemExit:
                load_error = load_error + 1
                LOGGER.exception(
                    "Error loading %s (plugin tried to exit)", name)
            else:
                try:
                    if plugin.has_setup():
                        plugin.setup(self)
                        # TODO: remove in Sopel 8
                        self.__setup_plugins_check_manual_url_callbacks(name)
                    plugin.register(self)
                except Exception as e:
                    load_error = load_error + 1
                    LOGGER.exception("Error in %s setup: %s", name, e)
                else:
                    load_success = load_success + 1
                    LOGGER.info("Plugin loaded: %s", name)

        total = sum([load_success, load_error, load_disabled])
        if total and load_success:
            LOGGER.info(
                "Registered %d plugins, %d failed, %d disabled",
                (load_success - 1),
                load_error,
                load_disabled)
        else:
            LOGGER.warning("Warning: Couldn't load any plugins")

    def __setup_plugins_check_manual_url_callbacks(self, name):
        # check if a plugin modified bot.memory['url_callbacks'] manually
        # TODO: remove in Sopel 8
        if 'url_callbacks' not in self.memory:
            # nothing to check
            return

        for key, callback in tools.iteritems(self.memory['url_callbacks']):
            is_checked = getattr(
                callback, '_sopel_url_callbacks_checked', False)
            if is_checked:
                # already checked; move on to next callback
                continue

            # deprecation warning
            LOGGER.warning(
                "Plugin `%s` uses `bot.memory['url_callbacks']`; "
                'this key is deprecated and will be removed in Sopel 8. '
                'Use `@url` or `@url_lazy` instead. Callback was: %s',
                name, callback.__name__)
            # mark callback as checked
            setattr(callback, '_sopel_url_callbacks_checked', True)

    # post setup

    def post_setup(self):
        """Perform post-setup actions.

        This method handles everything that should happen after all the plugins
        are loaded, and before the bot can connect to the IRC server.

        At the moment, this method checks for undefined configuration options,
        and starts the job scheduler.

        .. versionadded:: 7.1
        """
        settings = self.settings
        for section_name, section in settings.get_defined_sections():
            defined_options = {
                settings.parser.optionxform(opt)
                for opt, _ in inspect.getmembers(section)
                if not opt.startswith('_')
            }
            for option_name in settings.parser.options(section_name):
                if option_name not in defined_options:
                    LOGGER.warning(
                        "Config option `%s.%s` is not defined by its section "
                        "and may not be recognized by Sopel.",
                        section_name,
                        option_name,
                    )

        self._scheduler.start()

    # plugins management

    def reload_plugin(self, name):
        """Reload a plugin.

        :param str name: name of the plugin to reload
        :raise plugins.exceptions.PluginNotRegistered: when there is no
            ``name`` plugin registered

        This function runs the plugin's shutdown routine and unregisters the
        plugin from the bot. Then this function reloads the plugin, runs its
        setup routines, and registers it again.
        """
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        plugin = self._plugins[name]
        # tear down
        plugin.shutdown(self)
        plugin.unregister(self)
        LOGGER.info("Unloaded plugin %s", name)
        # reload & setup
        plugin.reload()
        plugin.setup(self)
        plugin.register(self)
        meta = plugin.get_meta_description()
        LOGGER.info("Reloaded %s plugin %s from %s",
                    meta['type'], name, meta['source'])

    def reload_plugins(self):
        """Reload all registered plugins.

        First, this function runs all plugin shutdown routines and unregisters
        all plugins. Then it reloads all plugins, runs their setup routines, and
        registers them again.
        """
        registered = list(self._plugins.items())
        # tear down all plugins
        for name, plugin in registered:
            plugin.shutdown(self)
            plugin.unregister(self)
            LOGGER.info("Unloaded plugin %s", name)

        # reload & setup all plugins
        for name, plugin in registered:
            plugin.reload()
            plugin.setup(self)
            plugin.register(self)
            meta = plugin.get_meta_description()
            LOGGER.info("Reloaded %s plugin %s from %s",
                        meta['type'], name, meta['source'])

    def add_plugin(self, plugin, callables, jobs, shutdowns, urls):
        """Add a loaded plugin to the bot's registry.

        :param plugin: loaded plugin to add
        :type plugin: :class:`sopel.plugins.handlers.AbstractPluginHandler`
        :param callables: an iterable of callables from the ``plugin``
        :type callables: :term:`iterable`
        :param jobs: an iterable of functions from the ``plugin`` that are
                     periodically invoked
        :type jobs: :term:`iterable`
        :param shutdowns: an iterable of functions from the ``plugin`` that
                          should be called on shutdown
        :type shutdowns: :term:`iterable`
        :param urls: an iterable of functions from the ``plugin`` to call when
                     matched against a URL
        :type urls: :term:`iterable`
        """
        self._plugins[plugin.name] = plugin
        self.register_callables(callables)
        self.register_jobs(jobs)
        self.register_shutdowns(shutdowns)
        self.register_urls(urls)

    def remove_plugin(self, plugin, callables, jobs, shutdowns, urls):
        """Remove a loaded plugin from the bot's registry.

        :param plugin: loaded plugin to remove
        :type plugin: :class:`sopel.plugins.handlers.AbstractPluginHandler`
        :param callables: an iterable of callables from the ``plugin``
        :type callables: :term:`iterable`
        :param jobs: an iterable of functions from the ``plugin`` that are
                     periodically invoked
        :type jobs: :term:`iterable`
        :param shutdowns: an iterable of functions from the ``plugin`` that
                          should be called on shutdown
        :type shutdowns: :term:`iterable`
        :param urls: an iterable of functions from the ``plugin`` to call when
                     matched against a URL
        :type urls: :term:`iterable`
        """
        name = plugin.name
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        # remove plugin rules, jobs, shutdown functions, and url callbacks
        self._rules_manager.unregister_plugin(name)
        self._scheduler.unregister_plugin(name)
        self.unregister_shutdowns(shutdowns)

        # remove plugin from registry
        del self._plugins[name]

    def has_plugin(self, name):
        """Check if the bot has registered a plugin of the specified name.

        :param str name: name of the plugin to check for
        :return: whether the bot has a plugin named ``name`` registered
        :rtype: bool
        """
        return name in self._plugins

    def get_plugin_meta(self, name):
        """Get info about a registered plugin by its name.

        :param str name: name of the plugin about which to get info
        :return: the plugin's metadata
                 (see :meth:`~.plugins.handlers.AbstractPluginHandler.get_meta_description`)
        :rtype: :class:`dict`
        :raise plugins.exceptions.PluginNotRegistered: when there is no
            ``name`` plugin registered
        """
        if not self.has_plugin(name):
            raise plugins.exceptions.PluginNotRegistered(name)

        return self._plugins[name].get_meta_description()

    # callable management

    @deprecated(
        reason="Replaced by specific `unregister_*` methods.",
        version='7.1',
        removed_in='8.0')
    def unregister(self, obj):
        """Unregister a shutdown method.

        :param obj: the shutdown method to unregister
        :type obj: :term:`object`

        This method was used to unregister anything (rules, commands, urls,
        jobs, and shutdown methods), but since everything can be done by other
        means, there is no use for it anymore.
        """
        callable_name = getattr(obj, "__name__", 'UNKNOWN')

        if hasattr(obj, 'interval'):
            self.unregister_jobs([obj])

        if callable_name == "shutdown" and obj in self.shutdown_methods:
            self.unregister_shutdowns([obj])

    @deprecated(
        reason="Replaced by specific `register_*` methods.",
        version='7.1',
        removed_in='8.0')
    def register(self, callables, jobs, shutdowns, urls):
        """Register rules, jobs, shutdown methods, and URL callbacks.

        :param callables: an iterable of callables to register
        :type callables: :term:`iterable`
        :param jobs: an iterable of functions to periodically invoke
        :type jobs: :term:`iterable`
        :param shutdowns: an iterable of functions to call on shutdown
        :type shutdowns: :term:`iterable`
        :param urls: an iterable of functions to call when matched against a URL
        :type urls: :term:`iterable`

        The ``callables`` argument contains a list of "callable objects", i.e.
        objects for which :func:`callable` will return ``True``. They can be:

        * a callable with rules (will match triggers with a regex pattern)
        * a callable without rules (will match any triggers, such as events)
        * a callable with commands
        * a callable with nick commands
        * a callable with action commands

        It is possible to have a callable with rules, commands, and nick
        commands configured. It should not be possible to have a callable with
        commands or nick commands but without rules.
        """
        self.register_callables(callables)
        self.register_jobs(jobs)
        self.register_shutdowns(shutdowns)
        self.register_urls(urls)

    def register_callables(self, callables):
        match_any = re.compile(r'.*')
        settings = self.settings

        for callbl in callables:
            rules = getattr(callbl, 'rule', [])
            lazy_rules = getattr(callbl, 'rule_lazy_loaders', [])
            find_rules = getattr(callbl, 'find_rules', [])
            lazy_find_rules = getattr(callbl, 'find_rules_lazy_loaders', [])
            search_rules = getattr(callbl, 'search_rules', [])
            lazy_search_rules = getattr(callbl, 'search_rules_lazy_loaders', [])
            commands = getattr(callbl, 'commands', [])
            nick_commands = getattr(callbl, 'nickname_commands', [])
            action_commands = getattr(callbl, 'action_commands', [])
            is_rule = any([
                rules,
                lazy_rules,
                find_rules,
                lazy_find_rules,
                search_rules,
                lazy_search_rules,
            ])
            is_command = any([commands, nick_commands, action_commands])

            if rules:
                rule = plugin_rules.Rule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_rules:
                try:
                    rule = plugin_rules.Rule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register rule: %s', err)

            if find_rules:
                rule = plugin_rules.FindRule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_find_rules:
                try:
                    rule = plugin_rules.FindRule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register find rule: %s', err)

            if search_rules:
                rule = plugin_rules.SearchRule.from_callable(settings, callbl)
                self._rules_manager.register(rule)

            if lazy_search_rules:
                try:
                    rule = plugin_rules.SearchRule.from_callable_lazy(
                        settings, callbl)
                    self._rules_manager.register(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error('Cannot register search rule: %s', err)

            if commands:
                rule = plugin_rules.Command.from_callable(settings, callbl)
                self._rules_manager.register_command(rule)

            if nick_commands:
                rule = plugin_rules.NickCommand.from_callable(
                    settings, callbl)
                self._rules_manager.register_nick_command(rule)

            if action_commands:
                rule = plugin_rules.ActionCommand.from_callable(
                    settings, callbl)
                self._rules_manager.register_action_command(rule)

            if not is_command and not is_rule:
                callbl.rule = [match_any]
                self._rules_manager.register(
                    plugin_rules.Rule.from_callable(self.settings, callbl))

    def register_jobs(self, jobs):
        for func in jobs:
            job = sopel.tools.jobs.Job.from_callable(self.settings, func)
            self._scheduler.register(job)

    def unregister_jobs(self, jobs):
        for job in jobs:
            self._scheduler.remove_callable_job(job)

    def register_shutdowns(self, shutdowns):
        # Append plugin's shutdown function to the bot's list of functions to
        # call on shutdown
        self.shutdown_methods = self.shutdown_methods + list(shutdowns)

    def unregister_shutdowns(self, shutdowns):
        self.shutdown_methods = [
            shutdown
            for shutdown in self.shutdown_methods
            if shutdown not in shutdowns
        ]

    def register_urls(self, urls):
        for func in urls:
            url_regex = getattr(func, 'url_regex', [])
            url_lazy_loaders = getattr(func, 'url_lazy_loaders', None)

            if url_regex:
                rule = plugin_rules.URLCallback.from_callable(
                    self.settings, func)
                self._rules_manager.register_url_callback(rule)

            if url_lazy_loaders:
                try:
                    rule = plugin_rules.URLCallback.from_callable_lazy(
                        self.settings, func)
                    self._rules_manager.register_url_callback(rule)
                except plugins.exceptions.PluginError as err:
                    LOGGER.error("Cannot register URL callback: %s", err)

    @deprecated(
        reason="Replaced by `say` method.",
        version='6.0',
        removed_in='8.0')
    def msg(self, recipient, text, max_messages=1):
        """Old way to make the bot say something on IRC.

        :param str recipient: nickname or channel to which to send message
        :param str text: message to send
        :param int max_messages: split ``text`` into at most this many messages
                                 if it is too long to fit in one (optional)

        .. deprecated:: 6.0
            Use :meth:`say` instead. Will be removed in Sopel 8.
        """
        self.say(text, recipient, max_messages)

    # message dispatch

    def call_rule(self, rule, sopel, trigger):
        # rate limiting
        if not trigger.admin and not rule.is_unblockable():
            if rule.is_rate_limited(trigger.nick):
                return
            if not trigger.is_privmsg and rule.is_channel_rate_limited(trigger.sender):
                return
            if rule.is_global_rate_limited():
                return

        # channel config
        if trigger.sender in self.config:
            channel_config = self.config[trigger.sender]

            # disable listed plugins completely on provided channel
            if 'disable_plugins' in channel_config:
                disabled_plugins = channel_config.disable_plugins.split(',')

                if '*' in disabled_plugins:
                    return
                elif rule.get_plugin_name() in disabled_plugins:
                    return

            # disable chosen methods from plugins
            if 'disable_commands' in channel_config:
                disabled_commands = literal_eval(channel_config.disable_commands)
                disabled_commands = disabled_commands.get(rule.get_plugin_name(), [])
                if rule.get_rule_label() in disabled_commands:
                    return

        try:
            rule.execute(sopel, trigger)
        except KeyboardInterrupt:
            raise
        except Exception as error:
            self.error(trigger, exception=error)

    def call(self, func, sopel, trigger):
        """Call a function, applying any rate limits or other restrictions.

        :param func: the function to call
        :type func: :term:`function`
        :param sopel: a SopelWrapper instance
        :type sopel: :class:`SopelWrapper`
        :param Trigger trigger: the Trigger object for the line from the server
                                that triggered this call
        """
        nick = trigger.nick
        current_time = time.time()
        if nick not in self._times:
            self._times[nick] = dict()
        if self.nick not in self._times:
            self._times[self.nick] = dict()
        if not trigger.is_privmsg and trigger.sender not in self._times:
            self._times[trigger.sender] = dict()

        if not trigger.admin and not func.unblockable:
            if func in self._times[nick]:
                usertimediff = current_time - self._times[nick][func]
                if func.rate > 0 and usertimediff < func.rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to user limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, usertimediff,
                        func.rate
                    )
                    return
            if func in self._times[self.nick]:
                globaltimediff = current_time - self._times[self.nick][func]
                if func.global_rate > 0 and globaltimediff < func.global_rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to global limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, globaltimediff,
                        func.global_rate
                    )
                    return

            if not trigger.is_privmsg and func in self._times[trigger.sender]:
                chantimediff = current_time - self._times[trigger.sender][func]
                if func.channel_rate > 0 and chantimediff < func.channel_rate:
                    LOGGER.info(
                        "%s prevented from using %s in %s due to channel limit: %d < %d",
                        trigger.nick, func.__name__, trigger.sender, chantimediff,
                        func.channel_rate
                    )
                    return

        # if channel has its own config section, check for excluded plugins/plugin methods
        if trigger.sender in self.config:
            channel_config = self.config[trigger.sender]
            LOGGER.debug(
                "Evaluating configuration for %s.%s in channel %s",
                func.plugin_name, func.__name__, trigger.sender
            )

            # disable listed plugins completely on provided channel
            if 'disable_plugins' in channel_config:
                disabled_plugins = channel_config.disable_plugins.split(',')

                # if "*" is used, we are disabling all plugins on provided channel
                if '*' in disabled_plugins:
                    LOGGER.debug(
                        "All plugins disabled in %s; skipping execution of %s.%s",
                        trigger.sender, func.plugin_name, func.__name__
                    )
                    return
                if func.plugin_name in disabled_plugins:
                    LOGGER.debug(
                        "Plugin %s is disabled in %s; skipping execution of %s",
                        func.plugin_name, trigger.sender, func.__name__
                    )
                    return

            # disable chosen methods from plugins
            if 'disable_commands' in channel_config:
                disabled_commands = literal_eval(channel_config.disable_commands)

                if func.plugin_name in disabled_commands:
                    if func.__name__ in disabled_commands[func.plugin_name]:
                        LOGGER.debug(
                            "Skipping execution of %s.%s in %s: disabled_commands matched",
                            func.plugin_name, func.__name__, trigger.sender
                        )
                        return

        try:
            exit_code = func(sopel, trigger)
        except Exception as error:  # TODO: Be specific
            exit_code = None
            self.error(trigger, exception=error)

        if exit_code != NOLIMIT:
            self._times[nick][func] = current_time
            self._times[self.nick][func] = current_time
            if not trigger.is_privmsg:
                self._times[trigger.sender][func] = current_time

    def _is_pretrigger_blocked(self, pretrigger):
        if self.settings.core.nick_blocks or self.settings.core.host_blocks:
            nick_blocked = self._nick_blocked(pretrigger.nick)
            host_blocked = self._host_blocked(pretrigger.host)
        else:
            nick_blocked = host_blocked = None

        return (nick_blocked, host_blocked)

    def dispatch(self, pretrigger):
        """Dispatch a parsed message to any registered callables.

        :param pretrigger: a parsed message from the server
        :type pretrigger: :class:`~sopel.trigger.PreTrigger`

        The ``pretrigger`` (a parsed message) is used to find matching rules;
        it will retrieve them by order of priority, and execute them. It runs
        triggered rules in separate threads, unless they are marked otherwise.

        However, it won't run triggered blockable rules at all when they can't
        be executed for blocked nickname or hostname.

        .. seealso::

            The pattern matching is done by the
            :class:`Rules Manager<sopel.plugins.rules.Manager>`.

        """
        # list of commands running in separate threads for this dispatch
        running_triggers = []
        # nickname/hostname blocking
        nick_blocked, host_blocked = self._is_pretrigger_blocked(pretrigger)
        blocked = bool(nick_blocked or host_blocked)
        list_of_blocked_rules = set()
        # account info
        nick = pretrigger.nick
        user_obj = self.users.get(nick)
        account = user_obj.account if user_obj else None

        for rule, match in self._rules_manager.get_triggered_rules(self, pretrigger):
            trigger = Trigger(self.settings, pretrigger, match, account)

            is_unblockable = trigger.admin or rule.is_unblockable()
            if blocked and not is_unblockable:
                list_of_blocked_rules.add(str(rule))
                continue

            wrapper = SopelWrapper(
                self, trigger, output_prefix=rule.get_output_prefix())

            if rule.is_threaded():
                # run in a separate thread
                targs = (rule, wrapper, trigger)
                t = threading.Thread(target=self.call_rule, args=targs)
                plugin_name = rule.get_plugin_name()
                rule_label = rule.get_rule_label()
                t.name = '%s-%s-%s' % (t.name, plugin_name, rule_label)
                t.start()
                running_triggers.append(t)
            else:
                # direct call
                self.call_rule(rule, wrapper, trigger)

        # update currently running triggers
        self._update_running_triggers(running_triggers)

        if list_of_blocked_rules:
            if nick_blocked and host_blocked:
                block_type = 'both blocklists'
            elif nick_blocked:
                block_type = 'nick blocklist'
            else:
                block_type = 'host blocklist'
            LOGGER.debug(
                "%s prevented from using %s by %s.",
                pretrigger.nick,
                ', '.join(list_of_blocked_rules),
                block_type,
            )

    @property
    def running_triggers(self):
        """Current active threads for triggers.

        :return: the running thread(s) currently processing trigger(s)
        :rtype: :term:`iterable`

        This is for testing and debugging purposes only.
        """
        with self._running_triggers_lock:
            return [t for t in self._running_triggers if t.is_alive()]

    def _update_running_triggers(self, running_triggers):
        """Update list of running triggers.

        :param list running_triggers: newly started threads

        We want to keep track of running triggers, mostly for testing and
        debugging purposes. For instance, it'll help make sure, in tests, that
        a bot plugin has finished processing a trigger, by manually joining
        all running threads.

        This is kept private, as it's purely internal machinery and isn't
        meant to be manipulated by outside code.
        """
        # update bot's global running triggers
        with self._running_triggers_lock:
            running_triggers = running_triggers + self._running_triggers
            self._running_triggers = [
                t for t in running_triggers if t.is_alive()]

    # event handlers

    def on_scheduler_error(self, scheduler, exc):
        """Called when the Job Scheduler fails.

        :param scheduler: the job scheduler that errored
        :type scheduler: :class:`sopel.plugins.jobs.Scheduler`
        :param Exception exc: the raised exception

        .. seealso::

            :meth:`Sopel.error`
        """
        self.error(exception=exc)

    def on_job_error(self, scheduler, job, exc):
        """Called when a job from the Job Scheduler fails.

        :param scheduler: the job scheduler responsible for the errored ``job``
        :type scheduler: :class:`sopel.plugins.jobs.Scheduler`
        :param job: the Job that errored
        :type job: :class:`sopel.tools.jobs.Job`
        :param Exception exc: the raised exception

        .. seealso::

            :meth:`Sopel.error`
        """
        self.error(exception=exc)

    def error(self, trigger=None, exception=None):
        """Called internally when a plugin causes an error.

        :param trigger: the ``Trigger``\\ing line (if available)
        :type trigger: :class:`sopel.trigger.Trigger`
        :param Exception exception: the exception raised by the error (if
                                    available)
        """
        message = 'Unexpected error'
        if exception:
            message = '{} ({})'.format(message, exception)

        if trigger:
            message = '{} from {} at {}. Message was: {}'.format(
                message, trigger.nick, str(datetime.utcnow()), trigger.group(0)
            )

        LOGGER.exception(message)

        if trigger and self.settings.core.reply_errors and trigger.sender is not None:
            self.say(message, trigger.sender)

    def _host_blocked(self, host):
        """Check if a hostname is blocked.

        :param str host: the hostname to check
        """
        bad_masks = self.config.core.host_blocks
        for bad_mask in bad_masks:
            bad_mask = bad_mask.strip()
            if not bad_mask:
                continue
            if (re.match(bad_mask + '$', host, re.IGNORECASE) or
                    bad_mask == host):
                return True
        return False

    def _nick_blocked(self, nick):
        """Check if a nickname is blocked.

        :param str nick: the nickname to check
        """
        bad_nicks = self.config.core.nick_blocks
        for bad_nick in bad_nicks:
            bad_nick = bad_nick.strip()
            if not bad_nick:
                continue
            if (re.match(bad_nick + '$', nick, re.IGNORECASE) or
                    Identifier(bad_nick) == nick):
                return True
        return False

    def _shutdown(self):
        """Internal bot shutdown method."""
        LOGGER.info("Shutting down")
        # Stop Job Scheduler
        LOGGER.info("Stopping the Job Scheduler.")
        self._scheduler.stop()

        try:
            self._scheduler.join(timeout=15)
        except RuntimeError:
            LOGGER.exception("Unable to stop the Job Scheduler.")
        else:
            LOGGER.info("Job Scheduler stopped.")

        self._scheduler.clear_jobs()

        # Shutdown plugins
        LOGGER.info(
            "Calling shutdown for %d plugins.", len(self.shutdown_methods))

        for shutdown_method in self.shutdown_methods:
            try:
                LOGGER.debug(
                    "Calling %s.%s",
                    shutdown_method.__module__,
                    shutdown_method.__name__)
                shutdown_method(self)
            except Exception as e:
                LOGGER.exception("Error calling shutdown method: %s", e)

        # Avoid calling shutdown methods if we already have.
        self.shutdown_methods = []

    # URL callbacks management

    @deprecated(
        reason='Issues with @url decorator have been fixed. Simply use that.',
        version='7.1',
        warning_in='8.0',
        removed_in='9.0',
    )
    def register_url_callback(self, pattern, callback):
        """Register a ``callback`` for URLs matching the regex ``pattern``.

        :param pattern: compiled regex pattern to register
        :type pattern: :ref:`re.Pattern <python:re-objects>`
        :param callback: callable object to handle matching URLs
        :type callback: :term:`function`

        .. versionadded:: 7.0

            This method replaces manual management of ``url_callbacks`` in
            Sopel's plugins, so instead of doing this in ``setup()``::

                if 'url_callbacks' not in bot.memory:
                    bot.memory['url_callbacks'] = tools.SopelMemory()

                regex = re.compile(r'http://example.com/path/.*')
                bot.memory['url_callbacks'][regex] = callback

            use this much more concise pattern::

                regex = re.compile(r'http://example.com/path/.*')
                bot.register_url_callback(regex, callback)

        It's recommended you completely avoid manual management of URL
        callbacks through the use of :func:`sopel.plugin.url`.

        .. deprecated:: 7.1

            Made obsolete by fixes to the behavior of
            :func:`sopel.plugin.url`. Will be removed in Sopel 9.

        """
        if 'url_callbacks' not in self.memory:
            self.memory['url_callbacks'] = tools.SopelMemory()

        if isinstance(pattern, basestring):
            pattern = re.compile(pattern)

        # Mark the callback as checked: using this method is safe.
        # TODO: remove in Sopel 8
        setattr(callback, '_sopel_url_callbacks_checked', True)
        self.memory['url_callbacks'][pattern] = callback

    @deprecated(
        reason='Issues with @url decorator have been fixed. Simply use that.',
        version='7.1',
        warning_in='8.0',
        removed_in='9.0',
    )
    def unregister_url_callback(self, pattern, callback):
        """Unregister the callback for URLs matching the regex ``pattern``.

        :param pattern: compiled regex pattern to unregister callback
        :type pattern: :ref:`re.Pattern <python:re-objects>`
        :param callback: callable object to remove
        :type callback: :term:`function`

        .. versionadded:: 7.0

            This method replaces manual management of ``url_callbacks`` in
            Sopel's plugins, so instead of doing this in ``shutdown()``::

                regex = re.compile(r'http://example.com/path/.*')
                try:
                    del bot.memory['url_callbacks'][regex]
                except KeyError:
                    pass

            use this much more concise pattern::

                regex = re.compile(r'http://example.com/path/.*')
                bot.unregister_url_callback(regex, callback)

        It's recommended you completely avoid manual management of URL
        callbacks through the use of :func:`sopel.plugin.url`.

        .. deprecated:: 7.1

            Made obsolete by fixes to the behavior of
            :func:`sopel.plugin.url`. Will be removed in Sopel 9.

        """
        if 'url_callbacks' not in self.memory:
            # nothing to unregister
            return

        if isinstance(pattern, basestring):
            pattern = re.compile(pattern)

        try:
            del self.memory['url_callbacks'][pattern]
        except KeyError:
            pass

    def search_url_callbacks(self, url):
        """Yield callbacks whose regex pattern matches the ``url``.

        :param str url: URL found in a trigger
        :return: yield 2-value tuples of ``(callback, match)``

        For each pattern that matches the ``url`` parameter, it yields a
        2-value tuple of ``(callable, match)`` for that pattern.

        The ``callable`` is the one registered with
        :meth:`register_url_callback`, and the ``match`` is the result of
        the regex pattern's ``search`` method.

        .. versionadded:: 7.0

        .. seealso::

            The Python documentation for the `re.search`__ function and
            the `match object`__.

        .. __: https://docs.python.org/3.6/library/re.html#re.search
        .. __: https://docs.python.org/3.6/library/re.html#match-objects

        """
        if 'url_callbacks' not in self.memory:
            # nothing to search
            return

        for regex, function in tools.iteritems(self.memory['url_callbacks']):
            match = regex.search(url)
            if match:
                yield function, match

    def restart(self, message):
        """Disconnect from IRC and restart the bot.

        :param str message: QUIT message to send (e.g. "Be right back!")
        """
        self.wantsrestart = True
        self.quit(message)


class SopelWrapper(object):
    """Wrapper around a Sopel instance and a Trigger.

    :param sopel: Sopel instance
    :type sopel: :class:`~sopel.bot.Sopel`
    :param trigger: IRC Trigger line
    :type trigger: :class:`~sopel.trigger.Trigger`
    :param str output_prefix: prefix for messages sent through this wrapper
                              (e.g. plugin tag)

    This wrapper will be used to call Sopel's triggered commands and rules as
    their ``bot`` argument. It acts as a proxy to :meth:`send messages<say>`
    to the sender (either a channel or in a private message) and even to
    :meth:`reply to someone<reply>` in a channel.
    """
    def __init__(self, sopel, trigger, output_prefix=''):
        if not output_prefix:
            # Just in case someone passes in False, None, etc.
            output_prefix = ''
        # The custom __setattr__ for this class sets the attribute on the
        # original bot object. We don't want that for these, so we set them
        # with the normal __setattr__.
        object.__setattr__(self, '_bot', sopel)
        object.__setattr__(self, '_trigger', trigger)
        object.__setattr__(self, '_out_pfx', output_prefix)

    def __dir__(self):
        classattrs = [attr for attr in self.__class__.__dict__
                      if not attr.startswith('__')]
        return list(self.__dict__) + classattrs + dir(self._bot)

    def __getattr__(self, attr):
        return getattr(self._bot, attr)

    def __setattr__(self, attr, value):
        return setattr(self._bot, attr, value)

    def say(self, message, destination=None, max_messages=1, truncation='', trailing=''):
        """Override ``Sopel.say`` to use trigger source by default.

        :param str message: message to say
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`
        :param int max_messages: split ``message`` into at most this many
                                 messages if it is too long to fit into one
                                 line (optional)
        :param str truncation: string to indicate that the ``message`` was
                               truncated (optional)
        :param str trailing: string that should always appear at the end of
                             ``message`` (optional)

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            For more details about the optional arguments to this wrapper
            method, consult the documentation for :meth:`sopel.bot.Sopel.say`.

        """
        if destination is None:
            destination = self._trigger.sender
        self._bot.say(self._out_pfx + message, destination, max_messages, truncation, trailing)

    def action(self, message, destination=None):
        """Override ``Sopel.action`` to use trigger source by default.

        :param str message: action message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            :meth:`sopel.bot.Sopel.action`
        """
        if destination is None:
            destination = self._trigger.sender
        self._bot.action(message, destination)

    def notice(self, message, destination=None):
        """Override ``Sopel.notice`` to use trigger source by default.

        :param str message: notice message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        .. seealso::

            :meth:`sopel.bot.Sopel.notice`
        """
        if destination is None:
            destination = self._trigger.sender
        self._bot.notice(self._out_pfx + message, destination)

    def reply(self, message, destination=None, reply_to=None, notice=False):
        """Override ``Sopel.reply`` to ``reply_to`` sender by default.

        :param str message: reply message
        :param str destination: channel or nickname; defaults to
            :attr:`trigger.sender <sopel.trigger.Trigger.sender>`
        :param str reply_to: person to reply to; defaults to
            :attr:`trigger.nick <sopel.trigger.Trigger.nick>`
        :param bool notice: reply as an IRC notice or with a simple message

        The ``destination`` will default to the channel in which the
        trigger happened (or nickname, if received in a private message).

        ``reply_to`` will default to the nickname who sent the trigger.

        .. seealso::

            :meth:`sopel.bot.Sopel.reply`
        """
        if destination is None:
            destination = self._trigger.sender
        if reply_to is None:
            reply_to = self._trigger.nick
        self._bot.reply(message, destination, reply_to, notice)

    def kick(self, nick, channel=None, message=None):
        """Override ``Sopel.kick`` to kick in a channel

        :param str nick: nick to kick out of the ``channel``
        :param str channel: optional channel to kick ``nick`` from
        :param str message: optional message for the kick

        The ``channel`` will default to the channel in which the call was
        triggered. If triggered from a private message, ``channel`` is
        required.

        .. seealso::

            :meth:`sopel.bot.Sopel.kick`
        """
        if channel is None:
            if self._trigger.is_privmsg:
                raise RuntimeError('Error: KICK requires a channel.')
            else:
                channel = self._trigger.sender
        if nick is None:
            raise RuntimeError('Error: KICK requires a nick.')
        self._bot.kick(nick, channel, message)



# coding=utf-8
"""Tests for the ``sopel.loader`` module."""
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
import re

import pytest

from sopel import loader, module, plugins


MOCK_MODULE_CONTENT = """# coding=utf-8
import re

import sopel.module
import sopel.plugin


@sopel.module.commands("first")
def first_command(bot, trigger):
    pass


@sopel.module.commands("second")
def second_command(bot, trigger):
    pass


@sopel.module.interval(5)
def interval5s(bot):
    pass


@sopel.module.interval(10)
def interval10s(bot):
    pass


@sopel.module.url(r'.\\.example\\.com')
def example_url(bot, trigger, match=None):
    pass


def loader(settings):
    return [re.compile(r'.+\\.example\\.com')]


@sopel.plugin.url_lazy(loader)
def example_url_lazy(bot, trigger):
    pass


@sopel.plugin.rule_lazy(loader)
def example_rule_lazy(bot, trigger):
    pass


@sopel.plugin.find_lazy(loader)
def example_find_lazy(bot, trigger):
    pass


@sopel.plugin.search_lazy(loader)
def example_search_lazy(bot, trigger):
    pass


@sopel.module.event('TOPIC')
def on_topic_command(bot):
    pass


def shutdown():
    pass


def ignored():
    pass


@sopel.module.rate(10)
def ignored_rate():
    pass


class Ignored:
    def __init__(self):
        self.rule = [r'.*']

    def __call__(self, bot, trigger):
        pass

ignored_obj = Ignored()

def ignored_trickster():
    pass

ignored_trickster._sopel_callable = True
"""


@pytest.fixture
def func():
    """Pytest fixture to get a function that will return True all the time"""
    def bot_command():
        """Test callable defined as a pytest fixture."""
        return True
    return bot_command


TMP_CONFIG = """
[core]
owner = testnick
nick = TestBot
"""


@pytest.fixture
def tmpconfig(configfactory):
    return configfactory('conf.ini', TMP_CONFIG)


@pytest.fixture
def testplugin(tmpdir):
    root = tmpdir.mkdir('loader_mods')
    mod_file = root.join('file_mod.py')
    mod_file.write(MOCK_MODULE_CONTENT)

    return plugins.handlers.PyFilePlugin(mod_file.strpath)


def test_is_limitable(testplugin):
    """Test is_limitable behavior before clean_module is called."""
    testplugin.load()
    test_mod = testplugin._module

    assert loader.is_limitable(test_mod.first_command)
    assert loader.is_limitable(test_mod.second_command)
    assert loader.is_limitable(test_mod.on_topic_command)

    assert not loader.is_limitable(test_mod.interval5s)
    assert not loader.is_limitable(test_mod.interval10s)
    assert not loader.is_limitable(test_mod.shutdown)

    assert loader.is_limitable(test_mod.example_url)
    assert loader.is_limitable(test_mod.example_url_lazy)

    assert loader.is_limitable(test_mod.example_rule_lazy)
    assert loader.is_limitable(test_mod.example_find_lazy)
    assert loader.is_limitable(test_mod.example_search_lazy)


def test_is_triggerable(testplugin):
    """Test is_triggerable behavior before clean_module is called."""
    testplugin.load()
    test_mod = testplugin._module

    assert loader.is_triggerable(test_mod.first_command)
    assert loader.is_triggerable(test_mod.second_command)
    assert loader.is_triggerable(test_mod.on_topic_command)

    assert not loader.is_triggerable(test_mod.interval5s)
    assert not loader.is_triggerable(test_mod.interval10s)

    assert not loader.is_triggerable(test_mod.shutdown)
    assert not loader.is_triggerable(test_mod.example_url)
    assert not loader.is_triggerable(test_mod.example_url_lazy)

    assert loader.is_triggerable(test_mod.example_rule_lazy)
    assert loader.is_triggerable(test_mod.example_find_lazy)
    assert loader.is_triggerable(test_mod.example_search_lazy)


def test_is_url_callback(testplugin):
    """Test is_triggerable behavior before clean_module is called."""
    testplugin.load()
    test_mod = testplugin._module

    assert not loader.is_url_callback(test_mod.first_command)
    assert not loader.is_url_callback(test_mod.second_command)
    assert not loader.is_url_callback(test_mod.on_topic_command)

    assert not loader.is_url_callback(test_mod.interval5s)
    assert not loader.is_url_callback(test_mod.interval10s)

    assert not loader.is_url_callback(test_mod.shutdown)

    assert loader.is_url_callback(test_mod.example_url)
    assert loader.is_url_callback(test_mod.example_url_lazy)

    assert not loader.is_url_callback(test_mod.example_rule_lazy)
    assert not loader.is_url_callback(test_mod.example_find_lazy)
    assert not loader.is_url_callback(test_mod.example_search_lazy)


def test_clean_module(testplugin, tmpconfig):
    testplugin.load()
    test_mod = testplugin._module

    callables, jobs, shutdowns, urls = loader.clean_module(
        test_mod, tmpconfig)

    assert len(callables) == 6
    assert test_mod.first_command in callables
    assert test_mod.second_command in callables
    assert test_mod.on_topic_command in callables
    assert test_mod.example_rule_lazy in callables
    assert test_mod.example_find_lazy in callables
    assert test_mod.example_search_lazy in callables
    assert len(jobs) == 2
    assert test_mod.interval5s in jobs
    assert test_mod.interval10s in jobs
    assert len(shutdowns)
    assert test_mod.shutdown in shutdowns
    assert len(urls) == 2
    assert test_mod.example_url in urls
    assert test_mod.example_url_lazy in urls

    # assert is_triggerable behavior *after* clean_module has been called
    assert loader.is_triggerable(test_mod.first_command)
    assert loader.is_triggerable(test_mod.second_command)
    assert loader.is_triggerable(test_mod.on_topic_command)
    assert loader.is_triggerable(test_mod.example_rule_lazy)
    assert loader.is_triggerable(test_mod.example_find_lazy)
    assert loader.is_triggerable(test_mod.example_search_lazy)

    assert not loader.is_triggerable(test_mod.interval5s)
    assert not loader.is_triggerable(test_mod.interval10s)

    assert not loader.is_triggerable(test_mod.shutdown)
    assert not loader.is_triggerable(test_mod.example_url)
    assert not loader.is_triggerable(test_mod.example_url_lazy)

    # ignored function is ignored
    assert test_mod.ignored not in callables
    assert test_mod.ignored not in jobs
    assert test_mod.ignored not in shutdowns
    assert test_mod.ignored not in urls
    # @rate doesn't create a callable and is ignored
    assert test_mod.ignored_rate not in callables
    assert test_mod.ignored_rate not in jobs
    assert test_mod.ignored_rate not in shutdowns
    assert test_mod.ignored_rate not in urls
    # object with a triggerable attribute are ignored by default
    assert loader.is_triggerable(test_mod.ignored_obj)
    assert test_mod.ignored_obj not in callables
    assert test_mod.ignored_obj not in jobs
    assert test_mod.ignored_obj not in shutdowns
    assert test_mod.ignored_obj not in urls
    # trickster function is ignored: it's still not a proper plugin callable
    assert not loader.is_triggerable(test_mod.ignored_trickster)
    assert test_mod.ignored_trickster not in callables
    assert test_mod.ignored_trickster not in jobs
    assert test_mod.ignored_trickster not in shutdowns
    assert test_mod.ignored_trickster not in urls


def test_clean_module_idempotency(testplugin, tmpconfig):
    testplugin.load()
    test_mod = testplugin._module

    callables, jobs, shutdowns, urls = loader.clean_module(
        test_mod, tmpconfig)

    # sanity assertions: check test_clean_module if any of these fails
    assert len(callables) == 6
    assert len(jobs) == 2
    assert len(shutdowns) == 1
    assert len(urls) == 2

    # recall clean_module, we should have the same result
    new_callables, new_jobs, new_shutdowns, new_urls = loader.clean_module(
        test_mod, tmpconfig)

    assert new_callables == callables
    assert new_jobs == jobs
    assert new_shutdowns == shutdowns
    assert new_urls == urls

    # assert is_triggerable behavior
    assert loader.is_triggerable(test_mod.first_command)
    assert loader.is_triggerable(test_mod.second_command)
    assert loader.is_triggerable(test_mod.on_topic_command)
    assert loader.is_triggerable(test_mod.example_rule_lazy)
    assert loader.is_triggerable(test_mod.example_find_lazy)
    assert loader.is_triggerable(test_mod.example_search_lazy)

    assert not loader.is_triggerable(test_mod.interval5s)
    assert not loader.is_triggerable(test_mod.interval10s)

    assert not loader.is_triggerable(test_mod.shutdown)
    assert not loader.is_triggerable(test_mod.example_url)


def test_clean_callable_default(tmpconfig, func):
    loader.clean_callable(func, tmpconfig)

    # Default values
    assert hasattr(func, 'thread')
    assert func.thread is True

    # Not added by default
    assert not hasattr(func, 'unblockable')
    assert not hasattr(func, 'priority')
    assert not hasattr(func, 'rate')
    assert not hasattr(func, 'channel_rate')
    assert not hasattr(func, 'global_rate')
    assert not hasattr(func, 'event')
    assert not hasattr(func, 'rule')
    assert not hasattr(func, 'find_rules')
    assert not hasattr(func, 'search_rules')
    assert not hasattr(func, 'rule_lazy_loaders')
    assert not hasattr(func, 'find_rules_lazy_loaders')
    assert not hasattr(func, 'search_rules_lazy_loaders')
    assert not hasattr(func, 'commands')
    assert not hasattr(func, 'nickname_commands')
    assert not hasattr(func, 'action_commands')
    assert not hasattr(func, 'intents')


def test_clean_callable_command(tmpconfig, func):
    setattr(func, 'commands', ['test'])
    loader.clean_callable(func, tmpconfig)

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0
    assert hasattr(func, 'event')
    assert func.event == ['PRIVMSG']
    assert not hasattr(func, 'rule')


def test_clean_callable_event(tmpconfig, func):
    setattr(func, 'event', ['low', 'UP', 'MiXeD'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['LOW', 'UP', 'MIXED']

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert func.event == ['LOW', 'UP', 'MIXED']

    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_event_string(tmpconfig, func):
    setattr(func, 'event', 'some')
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['SOME']

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert func.event == ['SOME']


def test_clean_callable_rule(tmpconfig, func):
    setattr(func, 'rule', [r'abc'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'rule')
    assert len(func.rule) == 1

    # Test the regex is compiled properly
    regex = re.compile(func.rule[0])
    assert regex.match('abc')
    assert regex.match('abcd')
    assert not regex.match('efg')

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.rule) == 1
    assert regex not in func.rule
    assert r'abc' in func.rule

    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_rule_string(tmpconfig, func):
    setattr(func, 'rule', r'abc')
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'rule')
    assert len(func.rule) == 1

    # Test the regex is compiled properly
    assert func.rule[0] == r'abc'

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.rule) == 1
    assert func.rule[0] == r'abc'


def test_clean_callable_rule_nick(tmpconfig, func):
    """Assert ``$nick`` in a rule is not replaced (deprecated feature)."""
    setattr(func, 'rule', [r'$nickhello'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'rule')
    assert len(func.rule) == 1

    # Test the regex is not compiled
    assert func.rule[0] == r'$nickhello'

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.rule) == 1
    assert func.rule[0] == r'$nickhello'


def test_clean_callable_rule_nickname(tmpconfig, func):
    """Assert ``$nickname`` in a rule is not replaced (deprecated feature)."""
    setattr(func, 'rule', [r'$nickname\s+hello'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'rule')
    assert len(func.rule) == 1

    # Test the regex is not compiled
    assert func.rule[0] == r'$nickname\s+hello'

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.rule) == 1
    assert func.rule[0] == r'$nickname\s+hello'


def test_clean_callable_find_rules(tmpconfig, func):
    setattr(func, 'find_rules', [r'abc'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'find_rules')
    assert len(func.find_rules) == 1
    assert not hasattr(func, 'rule')

    # Test the regex is compiled properly
    regex = re.compile(func.find_rules[0])
    assert regex.findall('abc')
    assert regex.findall('abcd')
    assert not regex.findall('adbc')

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert hasattr(func, 'find_rules')
    assert len(func.find_rules) == 1
    assert regex not in func.find_rules
    assert r'abc' in func.find_rules
    assert not hasattr(func, 'rule')

    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_search_rules(tmpconfig, func):
    setattr(func, 'search_rules', [r'abc'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'search_rules')
    assert len(func.search_rules) == 1
    assert not hasattr(func, 'rule')

    # Test the regex is compiled properly
    regex = re.compile(func.search_rules[0])
    assert regex.search('abc')
    assert regex.search('xyzabc')
    assert regex.search('abcd')
    assert not regex.search('adbc')

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert hasattr(func, 'search_rules')
    assert len(func.search_rules) == 1
    assert regex not in func.search_rules
    assert func.search_rules[0] == r'abc'
    assert not hasattr(func, 'rule')

    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_nickname_command(tmpconfig, func):
    setattr(func, 'nickname_commands', ['hello!'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'nickname_commands')
    assert len(func.nickname_commands) == 1
    assert func.nickname_commands == ['hello!']
    assert not hasattr(func, 'rule')

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert not hasattr(func, 'rule')
    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_action_command(tmpconfig, func):
    setattr(func, 'action_commands', ['bots'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'action_commands')
    assert len(func.action_commands) == 1
    assert func.action_commands == ['bots']
    assert not hasattr(func, 'rule')

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert not hasattr(func, 'rule')
    assert func.action_commands == ['bots']


def test_clean_callable_events(tmpconfig, func):
    setattr(func, 'event', ['TOPIC'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['TOPIC']

    setattr(func, 'event', ['TOPIC', 'JOIN'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['TOPIC', 'JOIN']

    setattr(func, 'event', ['TOPIC', 'join', 'Nick'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['TOPIC', 'JOIN', 'NICK']


def test_clean_callable_events_basestring(tmpconfig, func):
    setattr(func, 'event', 'topic')
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['TOPIC']

    setattr(func, 'event', 'JOIN')
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'event')
    assert func.event == ['JOIN']


def test_clean_callable_example(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello')(func)

    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, '_docs')
    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['.test hello']


def test_clean_callable_example_not_set(tmpconfig, func):
    module.commands('test')(func)

    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, '_docs')
    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == []


def test_clean_callable_example_multi_commands(tmpconfig, func):
    module.commands('test')(func)
    module.commands('unit')(func)
    module.example('.test hello')(func)

    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, '_docs')
    assert len(func._docs) == 2
    assert 'test' in func._docs
    assert 'unit' in func._docs

    test_docs = func._docs['test']
    unit_docs = func._docs['unit']
    assert len(test_docs) == 2
    assert test_docs == unit_docs

    assert test_docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert test_docs[1] == ['.test hello']


def test_clean_callable_example_first_only(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello')(func)
    module.example('.test bonjour')(func)

    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['.test hello']


def test_clean_callable_example_first_only_multi_commands(tmpconfig, func):
    module.commands('test')(func)
    module.commands('unit')(func)
    module.example('.test hello')(func)
    module.example('.test bonjour')(func)

    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, '_docs')
    assert len(func._docs) == 2
    assert 'test' in func._docs
    assert 'unit' in func._docs

    test_docs = func._docs['test']
    unit_docs = func._docs['unit']
    assert len(test_docs) == 2
    assert test_docs == unit_docs

    assert test_docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert test_docs[1] == ['.test hello']


def test_clean_callable_example_user_help(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello', user_help=True)(func)

    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['.test hello']


def test_clean_callable_example_user_help_multi(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello', user_help=True)(func)
    module.example('.test bonjour', user_help=True)(func)

    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['.test hello', '.test bonjour']


def test_clean_callable_example_user_help_mixed(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello')(func)
    module.example('.test bonjour', user_help=True)(func)

    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['.test bonjour']


def test_clean_callable_example_default_prefix(tmpconfig, func):
    module.commands('test')(func)
    module.example('.test hello')(func)

    tmpconfig.core.help_prefix = '!'
    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['!test hello']


def test_clean_callable_example_nickname(tmpconfig, func):
    module.commands('test')(func)
    module.example('$nickname: hello')(func)

    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['TestBot: hello']


def test_clean_callable_example_nickname_custom_prefix(tmpconfig, func):
    module.commands('test')(func)
    module.example('$nickname: hello')(func)

    tmpconfig.core.help_prefix = '!'
    loader.clean_callable(func, tmpconfig)

    assert len(func._docs) == 1
    assert 'test' in func._docs

    docs = func._docs['test']
    assert len(docs) == 2
    assert docs[0] == inspect.cleandoc(func.__doc__).splitlines()
    assert docs[1] == ['TestBot: hello']


def test_clean_callable_intents(tmpconfig, func):
    setattr(func, 'intents', [r'abc'])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'intents')
    assert len(func.intents) == 1

    # Test the regex is compiled properly
    regex = func.intents[0]
    assert regex.match('abc')
    assert regex.match('abcd')
    assert regex.match('ABC')
    assert regex.match('AbCdE')
    assert not regex.match('efg')

    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'priority')
    assert func.priority == 'medium'
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.intents) == 1
    assert regex in func.intents

    assert func.unblockable is False
    assert func.priority == 'medium'
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0


def test_clean_callable_url(tmpconfig, func):
    setattr(func, 'url_regex', [re.compile('.*')])
    loader.clean_callable(func, tmpconfig)

    assert hasattr(func, 'url_regex')
    assert len(func.url_regex) == 1

    # Don't test the regex; that's handled in a different module
    # Default values
    assert hasattr(func, 'unblockable')
    assert func.unblockable is False
    assert hasattr(func, 'thread')
    assert func.thread is True
    assert hasattr(func, 'rate')
    assert func.rate == 0
    assert hasattr(func, 'channel_rate')
    assert func.channel_rate == 0
    assert hasattr(func, 'global_rate')
    assert func.global_rate == 0

    # idempotency
    loader.clean_callable(func, tmpconfig)
    assert len(func.url_regex) == 1
    assert func.unblockable is False
    assert func.thread is True
    assert func.rate == 0
    assert func.channel_rate == 0
    assert func.global_rate == 0



# coding=utf-8
"""
seen.py - Sopel Seen Plugin
Copyright 2008, Sean B. Palmer, inamidst.com
Copyright © 2012, Elad Alfassa <elad@fedoraproject.org>
Copyright 2019, Sopel contributors
Licensed under the Eiffel Forum License 2.

https://sopel.chat
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import time

from sopel import plugin, tools
from sopel.tools.time import seconds_to_human


@plugin.command('seen')
@plugin.output_prefix('[seen] ')
def seen(bot, trigger):
    """Reports when and where the user was last seen."""
    if not trigger.group(2):
        bot.reply(
            "Use `%sseen <nick>` to know when <nick> was last seen."
            % bot.settings.core.help_prefix)
        return

    nick = trigger.group(2).strip()
    if nick == bot.nick:
        bot.reply("I'm right here!")
        return

    timestamp = bot.db.get_nick_value(nick, 'seen_timestamp')
    if not timestamp:
        bot.reply("Sorry, I haven't seen {nick} around.".format(nick=nick))
        return

    channel = bot.db.get_nick_value(nick, 'seen_channel')
    message = bot.db.get_nick_value(nick, 'seen_message')
    action = bot.db.get_nick_value(nick, 'seen_action')

    saw = datetime.datetime.utcfromtimestamp(timestamp)
    delta = seconds_to_human((trigger.time - saw).total_seconds())

    msg = "I last saw " + nick
    if tools.Identifier(channel) == trigger.sender:
        if action:
            msg += " in here {since}, doing: {nick} {action}".format(
                since=delta,
                nick=nick,
                action=message)
        else:
            msg += " in here {since}, saying: {message}".format(
                since=delta,
                message=message)
    else:
        msg += " in another channel {since}.".format(since=delta)
    bot.say(msg)


@plugin.thread(False)
@plugin.rule('(.*)')
@plugin.priority('low')
@plugin.unblockable
@plugin.require_chanmsg
def note(bot, trigger):
    nick = trigger.nick
    bot.db.set_nick_value(nick, 'seen_timestamp', time.time())
    bot.db.set_nick_value(nick, 'seen_channel', trigger.sender)
    bot.db.set_nick_value(nick, 'seen_message', trigger)
    bot.db.set_nick_value(nick, 'seen_action', trigger.ctcp is not None)



# coding=utf-8
# Copyright 2019, Florian Strzelecki <florian.strzelecki@gmail.com>
#
# Licensed under the Eiffel Forum License 2.
from __future__ import absolute_import, division, print_function, unicode_literals

from .utils import safe


class AbstractIRCBackend(object):
    """Abstract class defining the interface and basic logic of an IRC backend.

    :param bot: a Sopel instance
    :type bot: :class:`sopel.bot.Sopel`

    Some methods of this class **MUST** be overridden by a subclass, or the
    backend implementation will not function correctly.
    """
    def __init__(self, bot):
        self.bot = bot

    def is_connected(self):
        """Tell if the backend is connected or not.

        :rtype: bool
        """
        raise NotImplementedError

    def on_irc_error(self, pretrigger):
        """Action to perform when the server sends an error event.

        :param pretrigger: PreTrigger object with the error event
        :type pretrigger: :class:`sopel.trigger.PreTrigger`

        On IRC error, if ``bot.hasquit`` is set, the backend should close the
        connection so the bot can quit or reconnect as required.
        """
        raise NotImplementedError

    def irc_send(self, data):
        """Send an IRC line as raw ``data``.

        :param bytes data: raw line to send
        """
        raise NotImplementedError

    def send_command(self, *args, **kwargs):
        """Send a command through the IRC connection.

        :param args: IRC command to send with its argument(s)
        :param str text: the text to send (optional keyword argument)

        Example::

            # send the INFO command
            backend.send_command('INFO')
            # send the NICK command with the argument 'Sopel'
            backend.send_command('NICK', 'Sopel')
            # send the PRIVMSG command to channel #sopel with some text
            backend.send_command('PRIVMSG', '#sopel', text='Hello world!')

        .. note::

            This will call the :meth:`sopel.bot.Sopel.on_message_sent`
            callback on the bot instance with the raw message sent.
        """
        raw_command = self.prepare_command(*args, text=kwargs.get('text'))
        self.irc_send(raw_command.encode('utf-8'))
        self.bot.on_message_sent(raw_command)

    def prepare_command(self, *args, **kwargs):
        """Prepare an IRC command from ``args`` and optional ``text``.

        :param list args: list of text, arguments of the IRC command to send
        :param str text: optional text to send with the IRC command
        :return: the raw message to send through the connection
        :rtype: str

        From :rfc:`2812` Internet Relay Chat: Client Protocol, Section 2.3:

            IRC messages are always lines of characters terminated with a
            CR-LF (Carriage Return - Line Feed) pair, and these messages SHALL
            NOT exceed 512 characters in length, counting all characters
            including the trailing CR-LF. Thus, there are 510 characters
            maximum allowed for the command and its parameters. There is no
            provision for continuation of message lines.

        The length in the RFC refers to the length in *bytes*, which can be
        bigger than the length of the Unicode string. This method cuts the
        message until its length fits within this limit of 510 bytes.

        The returned message contains the CR-LF pair required at the end,
        and can be sent as-is.
        """
        text = kwargs.get('text')
        max_length = unicode_max_length = 510
        raw_command = ' '.join(args)
        if text is not None:
            raw_command = '{args} :{text}'.format(args=raw_command,
                                                  text=safe(text))

        # The max length of 512 is in bytes, not Unicode characters:
        # we can't split the message on bytes, or we may cut in the middle of a
        # multi-byte character.
        while len(raw_command.encode('utf-8')) > max_length:
            raw_command = raw_command[:unicode_max_length]
            unicode_max_length = unicode_max_length - 1

        # Ends the message with CR-LF
        return raw_command + '\r\n'

    def send_ping(self, host):
        """Send a ``PING`` command to the server.

        :param str host: IRC server host

        A ``PING`` command should be sent at a regular interval to make sure
        the server knows the IRC connection is still active.
        """
        self.send_command('PING', safe(host))

    def send_pong(self, host):
        """Send a ``PONG`` command to the server.

        :param str host: IRC server host

        A ``PONG`` command must be sent each time the server sends a ``PING``
        command to the client.
        """
        self.send_command('PONG', safe(host))

    def send_nick(self, nick):
        """Send a ``NICK`` command with a ``nick``.

        :param str nick: nickname to take
        """
        self.send_command('NICK', safe(nick))

    def send_user(self, user, mode, nick, name):
        """Send a ``USER`` command with a ``user``.

        :param str user: IRC username
        :param str mode: mode(s) to send for the user
        :param str nick: nickname associated with this user
        :param str name: "real name" for the user
        """
        self.send_command('USER', safe(user), mode, safe(nick), text=name)

    def send_pass(self, password):
        """Send a ``PASS`` command with a ``password``.

        :param str password: password for authentication
        """
        self.send_command('PASS', safe(password))

    def send_join(self, channel, password=None):
        """Send a ``JOIN`` command to ``channel`` with optional ``password``.

        :param str channel: channel to join
        :param str password: optional password for protected channels
        """
        if password is None:
            self.send_command('JOIN', safe(channel))
        else:
            self.send_command('JOIN', safe(channel), safe(password))

    def send_part(self, channel, reason=None):
        """Send a ``PART`` command to ``channel``.

        :param str channel: the channel to part
        :param str text: optional text for leaving the channel
        """
        self.send_command('PART', safe(channel), text=reason)

    def send_quit(self, reason=None):
        """Send a ``QUIT`` command.

        :param str reason: optional text for leaving the server

        This won't send anything if the backend isn't connected.
        """
        if self.is_connected():
            self.send_command('QUIT', text=reason)

    def send_kick(self, channel, nick, reason=None):
        """Send a ``KICK`` command for ``nick`` in ``channel`` .

        :param str channel: the channel from which to kick ``nick``
        :param str nick: nickname to kick from the ``channel``
        :param str reason: optional reason for the kick
        """
        self.send_command('KICK', safe(channel), safe(nick), text=reason)

    def send_privmsg(self, dest, text):
        """Send a ``PRIVMSG`` command to ``dest`` with ``text``.

        :param str dest: nickname or channel name
        :param str text: the text to send
        """
        self.send_command('PRIVMSG', safe(dest), text=text)

    def send_notice(self, dest, text):
        """Send a ``NOTICE`` command to ``dest`` with ``text``.

        :param str dest: nickname or channel name
        :param str text: the text to send
        """
        self.send_command('NOTICE', dest, text=text)
